import collections
from typing import Iterator

from strictdoc.core.document_tree import DocumentTree
from strictdoc.core.file_tree import File, FileOrFolderEntry, Folder


class DocumentTreeIterator:
    def __init__(self, document_tree: DocumentTree) -> None:
        assert isinstance(document_tree, DocumentTree)
        self.document_tree: DocumentTree = document_tree

    def is_empty_tree(self) -> bool:
        return len(self.document_tree.document_list) == 0

    def iterator_files_first(self) -> Iterator[FileOrFolderEntry]:
        task_list = collections.deque(
            map(
                lambda tree: tree.root_folder_or_file,
                self.document_tree.file_tree,
            )
        )

        while task_list:
            file_tree_or_file = task_list.popleft()
            if isinstance(file_tree_or_file, File):
                yield file_tree_or_file
            elif isinstance(file_tree_or_file, Folder):
                if not file_tree_or_file.has_sdoc_content:
                    continue
                yield file_tree_or_file
                task_list.extendleft(
                    reversed(file_tree_or_file.subfolder_trees)
                )
                task_list.extendleft(reversed(file_tree_or_file.files))
            else:  # pragma: no cover
                raise AssertionError



import os
import tempfile
from contextlib import ExitStack
from pathlib import Path
from typing import List, Tuple

import uvicorn

from strictdoc import __version__
from strictdoc.cli.cli_arg_parser import ServerCommandConfig
from strictdoc.core.project_config import ProjectConfig
from strictdoc.helpers.pickle import pickle_dump
from strictdoc.server.config import SDocServerEnvVariable
from strictdoc.server.reload_config import UvicornReloadConfig


def print_warning_message() -> None:
    strictdoc_version = "StrictDoc web server v" + __version__
    print(  # noqa: T201
        f"""
*********************************************************
* {strictdoc_version.center(53)} *
*                                                       *
* Share feedback and report issues on GitHub:           *
* https://github.com/strictdoc-project/strictdoc/issues *
*********************************************************
""",
        flush=True,
    )


def run_strictdoc_server(
    *, server_config: ServerCommandConfig, project_config: ProjectConfig
) -> None:
    print_warning_message()

    with ExitStack() as stack:
        if server_config.output_path is None:
            server_config.output_path = stack.enter_context(
                tempfile.TemporaryDirectory()
            )

        project_config.integrate_server_config(server_config)

        reload_config = UvicornReloadConfig.create(
            project_config, server_config
        )

        # The server's uvicorn.run does not support passing arguments to the
        # main function (strictdoc_production_app). Passing the pickled config
        # through the environmental variables interface.
        tmp_config_file = stack.enter_context(tempfile.NamedTemporaryFile())
        config_dump = pickle_dump(project_config)
        tmp_config_file.write(config_dump)
        tmp_config_file.flush()

        os.environ[SDocServerEnvVariable.PATH_TO_CONFIG] = tmp_config_file.name

        # The uvicorn config code uses this function to additionally resolve
        # the reload includes and excludes. Unfortunately, it does not work
        # correctly with the expanded globs that are created by StrictDoc.
        # Not doing any resolutions seems to work fine with StrictDoc, so using
        # a 'do nothing' stub and an assert below to make sure that this function
        # is not removed by uvicorn.
        def dont_resolve_reload_patterns(
            patterns_list: List[str], directories_list: List[str]
        ) -> Tuple[List[str], List[Path]]:  # pragma: no cover
            return patterns_list, list(  # pragma: no cover
                map(Path, directories_list)
            )

        assert hasattr(uvicorn.config, "resolve_reload_patterns"), (
            "REGRESSION: The function 'resolve_reload_patterns' is not defined "
            "in uvicorn.config. Please report this to StrictDoc developers at "
            "https://github.com/strictdoc-project/strictdoc/issues."
        )
        uvicorn.config.resolve_reload_patterns = dont_resolve_reload_patterns

        uvicorn.run(
            "strictdoc.server.app:strictdoc_production_app",
            app_dir=".",
            factory=True,
            host=server_config.host
            if server_config.host is not None
            else project_config.server_host,
            log_level="info",
            # "server" command port overrides the strictdoc.toml option for now.
            # Eventually, I am considering to remove the CLI option for the
            # port, to simplify the maintenance of two configs: the config file
            # and the "export" command's interface.
            port=(
                server_config.port
                if server_config.port is not None
                else project_config.server_port
            ),
            reload=reload_config.reload,
            reload_dirs=reload_config.reload_dirs,
            reload_includes=reload_config.reload_includes,
            reload_excludes=reload_config.reload_excludes,
        )



"""
@relation(SDOC-SRS-72, scope=file)
"""

import datetime
import uuid
from collections import defaultdict
from enum import Enum
from typing import Dict, List, Optional, Tuple, Union

from reqif.helpers.string.xhtml_indent import reqif_indent_xhtml_string
from reqif.models.reqif_core_content import ReqIFCoreContent
from reqif.models.reqif_data_type import (
    ReqIFDataTypeDefinitionEnumeration,
    ReqIFDataTypeDefinitionString,
    ReqIFDataTypeDefinitionXHTML,
    ReqIFEnumValue,
)
from reqif.models.reqif_namespace_info import ReqIFNamespaceInfo
from reqif.models.reqif_req_if_content import ReqIFReqIFContent
from reqif.models.reqif_reqif_header import ReqIFReqIFHeader
from reqif.models.reqif_spec_hierarchy import ReqIFSpecHierarchy
from reqif.models.reqif_spec_object import ReqIFSpecObject, SpecObjectAttribute
from reqif.models.reqif_spec_object_type import (
    ReqIFSpecObjectType,
    SpecAttributeDefinition,
)
from reqif.models.reqif_spec_relation import ReqIFSpecRelation
from reqif.models.reqif_spec_relation_type import ReqIFSpecRelationType
from reqif.models.reqif_specification import ReqIFSpecification
from reqif.models.reqif_specification_type import ReqIFSpecificationType
from reqif.models.reqif_types import SpecObjectAttributeType
from reqif.object_lookup import ReqIFObjectLookup
from reqif.reqif_bundle import ReqIFBundle

from strictdoc.backend.reqif.sdoc_reqif_fields import (
    SDOC_SPECIFICATION_TYPE_SINGLETON,
    map_sdoc_field_title_to_reqif_field_title,
)
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.document_grammar import DocumentGrammar
from strictdoc.backend.sdoc.models.grammar_element import (
    GrammarElementField,
    GrammarElementFieldMultipleChoice,
    GrammarElementFieldSingleChoice,
    GrammarElementFieldString,
)
from strictdoc.backend.sdoc.models.node import SDocNode
from strictdoc.backend.sdoc.models.reference import (
    ChildReqReference,
    ParentReqReference,
    Reference,
)
from strictdoc.backend.sdoc.models.section import SDocSection
from strictdoc.core.document_iterator import DocumentCachingIterator
from strictdoc.core.document_tree import DocumentTree
from strictdoc.helpers.cast import assert_cast
from strictdoc.helpers.ordered_set import OrderedSet
from strictdoc.helpers.string import escape


class StrictDocReqIFTypes(Enum):
    SINGLE_LINE_STRING = "SDOC_DATATYPE_SINGLE_LINE_STRING"
    MULTI_LINE_STRING = "SDOC_DATATYPE_MULTI_LINE_STRING"
    SINGLE_CHOICE = "SDOC_DATATYPE_SINGLE_CHOICE"
    MULTI_CHOICE = "SDOC_DATATYPE_MULTI_CHOICE"


REQIF_SINGLELINE_MAX_LENGTH = "10000"


def generate_unique_identifier(element_type: str) -> str:
    return f"{element_type}-{uuid.uuid4()}"


class P01_SDocToReqIFBuildContext:
    def __init__(self, *, multiline_is_xhtml: bool, enable_mid: bool):
        self.multiline_is_xhtml: bool = multiline_is_xhtml
        self.enable_mid: bool = enable_mid
        self.map_uid_to_spec_objects: Dict[str, ReqIFSpecObject] = {}
        self.map_node_uids_to_their_relations: Dict[str, List[Reference]] = (
            defaultdict(list)
        )

        # Maps SDoc field titles to ReqiF Data Type identifiers.
        self.data_types_lookup: Dict[str, str] = {}

        self.map_grammar_node_tags_to_spec_object_type: Dict[
            SDocDocument, Dict[str, ReqIFSpecObjectType]
        ] = defaultdict(dict)
        self.map_spec_relation_tuple_to_spec_relation_type: Dict[
            Tuple[str, Optional[str]], ReqIFSpecRelationType
        ] = {}
        self.export_date_str: str = datetime.datetime.now(
            datetime.timezone.utc
        ).strftime("%Y-%m-%dT%H:%M:%SZ")


class P01_SDocToReqIFObjectConverter:
    @classmethod
    def convert_document_tree(
        cls,
        document_tree: DocumentTree,
        multiline_is_xhtml: bool,
        enable_mid: bool,
    ) -> ReqIFBundle:
        creation_time = datetime.datetime.now(
            datetime.datetime.now().astimezone().tzinfo
        ).isoformat()

        namespace = "http://www.omg.org/spec/ReqIF/20110401/reqif.xsd"

        context: P01_SDocToReqIFBuildContext = P01_SDocToReqIFBuildContext(
            multiline_is_xhtml=multiline_is_xhtml, enable_mid=enable_mid
        )

        spec_types: List[ReqIFSpecificationType] = []
        spec_objects: List[ReqIFSpecObject] = []
        spec_relations: List[ReqIFSpecRelation] = []
        specifications: List[ReqIFSpecification] = []
        data_types: List[
            Union[
                ReqIFDataTypeDefinitionString,
                ReqIFDataTypeDefinitionXHTML,
                ReqIFDataTypeDefinitionEnumeration,
            ]
        ] = []

        specification_type = ReqIFSpecificationType(
            identifier=SDOC_SPECIFICATION_TYPE_SINGLETON,
            description=None,
            last_change=creation_time,
            long_name=SDOC_SPECIFICATION_TYPE_SINGLETON,
            spec_attributes=None,
            spec_attribute_map={},
            is_self_closed=True,
        )
        spec_types.append(specification_type)

        document: SDocDocument
        for document in document_tree.document_list:
            assert document.grammar is not None
            for element in document.grammar.elements:
                for field in element.fields:
                    multiline = element.is_field_multiline(field.title)

                    if isinstance(field, GrammarElementFieldString):
                        if multiline:
                            if (
                                StrictDocReqIFTypes.MULTI_LINE_STRING.value
                                in context.data_types_lookup
                            ):
                                continue
                            if multiline_is_xhtml:
                                data_type = ReqIFDataTypeDefinitionXHTML(
                                    identifier=(
                                        StrictDocReqIFTypes.MULTI_LINE_STRING.value
                                    ),
                                    last_change=context.export_date_str,
                                    is_self_closed=True,
                                )
                            else:
                                data_type = ReqIFDataTypeDefinitionString(
                                    identifier=(
                                        StrictDocReqIFTypes.MULTI_LINE_STRING.value
                                    ),
                                    last_change=context.export_date_str,
                                    max_length=REQIF_SINGLELINE_MAX_LENGTH,
                                )
                            context.data_types_lookup[
                                StrictDocReqIFTypes.MULTI_LINE_STRING.value
                            ] = data_type.identifier
                        else:
                            if (
                                StrictDocReqIFTypes.SINGLE_LINE_STRING.value
                                in context.data_types_lookup
                            ):
                                continue
                            data_type = ReqIFDataTypeDefinitionString(
                                identifier=(
                                    StrictDocReqIFTypes.SINGLE_LINE_STRING.value
                                ),
                                last_change=context.export_date_str,
                                max_length=REQIF_SINGLELINE_MAX_LENGTH,
                            )
                            context.data_types_lookup[
                                StrictDocReqIFTypes.SINGLE_LINE_STRING.value
                            ] = data_type.identifier
                        data_types.append(data_type)
                    elif isinstance(field, GrammarElementFieldSingleChoice):
                        values = []
                        values_map = {}
                        for option_idx_, option in enumerate(field.options):
                            value = ReqIFEnumValue(
                                identifier=generate_unique_identifier(
                                    "ENUM-VALUE"
                                ),
                                key=str(option_idx_),
                                last_change=context.export_date_str,
                                # ReqIF XML validator wants OTHER-CONTENT to be
                                # present, even if empty.
                                other_content="",
                                long_name=option,
                            )
                            values.append(value)
                            values_map[option] = option

                        data_type = ReqIFDataTypeDefinitionEnumeration(
                            identifier=(
                                generate_unique_identifier(
                                    StrictDocReqIFTypes.SINGLE_CHOICE.value
                                )
                            ),
                            last_change=context.export_date_str,
                            values=values,
                        )
                        data_types.append(data_type)
                        context.data_types_lookup[field.title] = (
                            data_type.identifier
                        )
                    elif isinstance(field, GrammarElementFieldMultipleChoice):
                        values = []
                        values_map = {}
                        for option_idx_, option in enumerate(field.options):
                            value = ReqIFEnumValue(
                                identifier=generate_unique_identifier(
                                    "ENUM-VALUE"
                                ),
                                key=str(option_idx_),
                                last_change=context.export_date_str,
                                # ReqIF XML validator wants OTHER-CONTENT to be
                                # present, even if empty.
                                other_content="",
                                long_name=option,
                            )
                            values.append(value)
                            values_map[option] = option

                        data_type = ReqIFDataTypeDefinitionEnumeration(
                            identifier=(
                                generate_unique_identifier(
                                    StrictDocReqIFTypes.MULTI_CHOICE.value
                                )
                            ),
                            last_change=context.export_date_str,
                            values=values,
                        )
                        data_types.append(data_type)
                        context.data_types_lookup[field.title] = (
                            data_type.identifier
                        )
                    else:
                        raise NotImplementedError(  # pragma: no cover
                            field
                        ) from None

            document_spec_types = cls._convert_document_grammar_to_spec_types(
                grammar=assert_cast(document.grammar, DocumentGrammar),
                context=context,
            )
            spec_types.extend(document_spec_types)

            document_iterator = DocumentCachingIterator(document)

            # FIXME: This is a throw-away object. It gets discarded when the
            #        iteration is over. Find a way to do without it.
            root_hierarchy = ReqIFSpecHierarchy(
                xml_node=None,
                is_self_closed=False,
                identifier="NOT_USED",
                last_change=None,
                long_name=None,
                spec_object="NOT_USED",
                children=[],
                ref_then_children_order=True,
                level=0,
            )

            node_stack: List[ReqIFSpecHierarchy] = [root_hierarchy]

            # FIXME: ReqIF must export complete documents including fragments.
            for node_, _ in document_iterator.all_content(
                print_fragments=False
            ):
                if isinstance(node_, SDocSection):
                    raise AssertionError(
                        "[SECTION] tags are deprecated when using ReqIF export/import. "
                        "Use [[SECTION]] tags instead. "
                        "See the migration guide for more details: "
                        "https://strictdoc.readthedocs.io/en/latest/latest/docs/strictdoc_01_user_guide.html#SECTION-UG-NODE-MIGRATION"
                    )

                if not isinstance(node_, SDocNode):
                    continue
                leaf_or_composite_node = assert_cast(node_, SDocNode)
                while len(node_stack) > assert_cast(node_.ng_level, int):
                    node_stack.pop()

                current_hierarchy = node_stack[-1]

                spec_object = cls._convert_requirement_to_spec_object(
                    requirement=leaf_or_composite_node,
                    grammar=assert_cast(document.grammar, DocumentGrammar),
                    context=context,
                    data_types=data_types,
                )
                spec_objects.append(spec_object)
                hierarchy = ReqIFSpecHierarchy(
                    xml_node=None,
                    is_self_closed=False,
                    identifier=generate_unique_identifier("SPEC-IDENTIFIER"),
                    last_change=context.export_date_str,
                    long_name=None,
                    spec_object=spec_object.identifier,
                    children=None,
                    ref_then_children_order=True,
                    level=leaf_or_composite_node.ng_level,
                )
                current_hierarchy.add_child(hierarchy)
                if leaf_or_composite_node.is_composite:
                    node_stack.append(hierarchy)

            specification_identifier: str
            if context.enable_mid and document.reserved_mid is not None:
                specification_identifier = document.reserved_mid
            else:
                specification_identifier = generate_unique_identifier(
                    "SPECIFICATION"
                )
            specification = ReqIFSpecification(
                xml_node=None,
                description=None,
                identifier=specification_identifier,
                last_change=context.export_date_str,
                long_name=document.title,
                values=None,
                specification_type=specification_type.identifier,
                children=root_hierarchy.children,
            )
            specifications.append(specification)

        for (
            requirement_id,
            node_relations_,
        ) in context.map_node_uids_to_their_relations.items():
            spec_object = context.map_uid_to_spec_objects[requirement_id]
            for node_relation_ in node_relations_:
                # For now, the File-relations are not supported.
                if not isinstance(
                    node_relation_, (ParentReqReference, ChildReqReference)
                ):
                    continue
                parent_or_child_relation = assert_cast(
                    node_relation_, (ParentReqReference, ChildReqReference)
                )
                related_node_uid = parent_or_child_relation.ref_uid
                parent_spec_object = context.map_uid_to_spec_objects[
                    related_node_uid
                ]
                spec_relation_type: ReqIFSpecRelationType = (
                    context.map_spec_relation_tuple_to_spec_relation_type[
                        (node_relation_.ref_type, node_relation_.role)
                    ]
                )
                spec_relations.append(
                    ReqIFSpecRelation(
                        xml_node=None,
                        description=None,
                        identifier=generate_unique_identifier("SPEC-RELATION"),
                        last_change=context.export_date_str,
                        relation_type_ref=spec_relation_type.identifier,
                        source=spec_object.identifier,
                        target=parent_spec_object.identifier,
                        values_attribute=None,
                    )
                )

        reqif_reqif_content = ReqIFReqIFContent(
            data_types=data_types,
            spec_types=spec_types,
            spec_objects=spec_objects,
            spec_relations=spec_relations,
            specifications=specifications,
            spec_relation_groups=None,
        )
        core_content_or_none = ReqIFCoreContent(reqif_reqif_content)

        namespace_info: ReqIFNamespaceInfo = ReqIFNamespaceInfo(
            original_reqif_tag_dump=None,
            doctype_is_present=True,
            encoding="UTF-8",
            namespace=namespace,
            configuration=None,
            namespace_id=None,
            namespace_xhtml="http://www.w3.org/1999/xhtml",
            schema_namespace=None,
            schema_location=None,
            language=None,
        )
        req_reqif_header = ReqIFReqIFHeader(
            identifier=generate_unique_identifier("REQ-IF-HEADER"),
            creation_time=creation_time,
            title="Documentation export by StrictDoc",
            req_if_tool_id="strictdoc",
            req_if_version="1.0",
            source_tool_id="strictdoc",
            repository_id=None,
            comment=None,
        )

        reqif_bundle = ReqIFBundle(
            namespace_info=namespace_info,
            req_if_header=req_reqif_header,
            core_content=core_content_or_none,
            tool_extensions_tag_exists=False,
            lookup=ReqIFObjectLookup(
                data_types_lookup={},
                spec_types_lookup={},
                spec_objects_lookup={},
                spec_relations_parent_lookup={},
            ),
            exceptions=[],
        )
        return reqif_bundle

    @classmethod
    def _convert_requirement_to_spec_object(
        cls,
        requirement: SDocNode,
        grammar: DocumentGrammar,
        context: P01_SDocToReqIFBuildContext,
        data_types: List[ReqIFDataTypeDefinitionString],
    ) -> ReqIFSpecObject:
        node_document = assert_cast(requirement.get_document(), SDocDocument)

        spec_object_type: ReqIFSpecObjectType = (
            context.map_grammar_node_tags_to_spec_object_type[node_document][
                requirement.node_type
            ]
        )

        enable_mid = context.enable_mid and node_document.config.enable_mid

        requirement_identifier: str
        if enable_mid and requirement.reserved_mid is not None:
            requirement_identifier = requirement.reserved_mid
        else:
            requirement_identifier = generate_unique_identifier(
                requirement.node_type
            )

        grammar_element = grammar.elements_by_type[requirement.node_type]

        attributes: List[SpecObjectAttribute] = []
        for field in requirement.fields_as_parsed:
            # The MID field, if exists, is extracted separately as a ReqIF Identifier.
            if field.field_name == "MID":
                continue

            field_name = map_sdoc_field_title_to_reqif_field_title(
                field.field_name,
                grammar_element.property_is_composite == True,
            )

            field_identifier = spec_object_type.identifier + "_" + field_name

            grammar_field = grammar_element.fields_map[field.field_name]
            if isinstance(grammar_field, GrammarElementFieldSingleChoice):
                data_type_ref = context.data_types_lookup[field.field_name]

                enum_ref_value = None
                for data_type in data_types:
                    if data_type_ref == data_type.identifier:
                        assert isinstance(
                            data_type, ReqIFDataTypeDefinitionEnumeration
                        )
                        for data_type_value in data_type.values:
                            if data_type_value.long_name is not None:
                                data_type_sdoc_value = data_type_value.long_name
                            else:
                                data_type_sdoc_value = data_type_value.key
                            if data_type_sdoc_value == field.get_text_value():
                                enum_ref_value = data_type_value.identifier
                                break

                assert enum_ref_value is not None

                attribute = SpecObjectAttribute(
                    xml_node=None,
                    attribute_type=SpecObjectAttributeType.ENUMERATION,
                    definition_ref=field_identifier,
                    value=[enum_ref_value],
                )
            elif isinstance(grammar_field, GrammarElementFieldMultipleChoice):
                field_values: List[str] = field.get_text_value().split(",")
                field_values = list(map(lambda v: v.strip(), field_values))

                data_type_ref = context.data_types_lookup[field.field_name]

                data_type_lookup = {}
                for data_type in data_types:
                    if data_type_ref == data_type.identifier:
                        assert isinstance(
                            data_type, ReqIFDataTypeDefinitionEnumeration
                        )
                        for data_type_value in data_type.values:
                            if data_type_value.long_name is not None:
                                data_type_sdoc_value = data_type_value.long_name
                            else:
                                data_type_sdoc_value = data_type_value.key
                            data_type_lookup[data_type_sdoc_value] = (
                                data_type_value.identifier
                            )

                field_values_refs = []
                for field_value_ in field_values:
                    field_values_refs.append(data_type_lookup[field_value_])

                attribute = SpecObjectAttribute(
                    xml_node=None,
                    attribute_type=SpecObjectAttributeType.ENUMERATION,
                    definition_ref=field_identifier,
                    value=field_values_refs,
                )
            elif isinstance(grammar_field, GrammarElementFieldString):
                is_multiline_field = field.is_multiline()

                field_value: str = field.get_text_value().rstrip()

                attribute_type: SpecObjectAttributeType
                if is_multiline_field:
                    if context.multiline_is_xhtml:
                        attribute_type = SpecObjectAttributeType.XHTML
                        field_value = (
                            "<xhtml:div>\n  "
                            + "\n  ".join(
                                f"<xhtml:p>{line}</xhtml:p>"
                                for line in field_value.split("\n\n")
                            )
                            + "\n</xhtml:div>"
                        )
                        field_value = reqif_indent_xhtml_string(field_value)
                    else:
                        attribute_type = SpecObjectAttributeType.STRING
                        field_value = escape(field_value)
                else:
                    field_value = escape(field_value)
                    attribute_type = SpecObjectAttributeType.STRING

                attribute = SpecObjectAttribute(
                    xml_node=None,
                    attribute_type=attribute_type,
                    definition_ref=field_identifier,
                    value=field_value,
                )
            else:
                raise NotImplementedError(  # pragma: no cover
                    grammar_field
                ) from None
            attributes.append(attribute)

        if requirement.reserved_uid is not None:
            if len(requirement.relations) > 0:
                context.map_node_uids_to_their_relations[
                    requirement.reserved_uid
                ] = requirement.relations

        spec_object = ReqIFSpecObject(
            identifier=requirement_identifier,
            spec_object_type=spec_object_type.identifier,
            attributes=attributes,
            last_change=context.export_date_str,
        )
        if requirement.reserved_uid is not None:
            context.map_uid_to_spec_objects[requirement.reserved_uid] = (
                spec_object
            )
        return spec_object

    @classmethod
    def _convert_document_grammar_to_spec_types(
        cls,
        grammar: DocumentGrammar,
        context: P01_SDocToReqIFBuildContext,
    ) -> List[ReqIFSpecificationType]:
        spec_object_types: List[ReqIFSpecificationType] = []

        grammar_document: SDocDocument = assert_cast(
            grammar.parent, SDocDocument
        )

        for element in grammar.elements:
            spec_object_type_identifier = element.tag + "_" + uuid.uuid4().hex

            attribute_definitions = []

            field: GrammarElementField
            for field in element.fields:
                multiline = element.is_field_multiline(field.title)
                field_title = map_sdoc_field_title_to_reqif_field_title(
                    field.title, element.property_is_composite == True
                )
                field_human_title = (
                    field.human_title
                    if field.human_title is not None
                    else field_title
                )
                field_identifier = (
                    spec_object_type_identifier + "_" + field_title
                )

                if isinstance(field, GrammarElementFieldString):
                    if multiline:
                        attribute_type = (
                            SpecObjectAttributeType.XHTML
                            if context.multiline_is_xhtml
                            else SpecObjectAttributeType.STRING
                        )
                        attribute = SpecAttributeDefinition(
                            attribute_type=attribute_type,
                            identifier=field_identifier,
                            datatype_definition=(
                                StrictDocReqIFTypes.MULTI_LINE_STRING.value
                            ),
                            long_name=field_human_title,
                            last_change=context.export_date_str,
                        )
                    else:
                        attribute = SpecAttributeDefinition(
                            attribute_type=SpecObjectAttributeType.STRING,
                            identifier=field_identifier,
                            datatype_definition=(
                                StrictDocReqIFTypes.SINGLE_LINE_STRING.value
                            ),
                            long_name=field_human_title,
                            last_change=context.export_date_str,
                        )
                elif isinstance(field, GrammarElementFieldSingleChoice):
                    attribute = SpecAttributeDefinition(
                        attribute_type=SpecObjectAttributeType.ENUMERATION,
                        identifier=field_identifier,
                        datatype_definition=context.data_types_lookup[
                            field.title
                        ],
                        long_name=field_human_title,
                        multi_valued=False,
                        last_change=context.export_date_str,
                    )
                elif isinstance(field, GrammarElementFieldMultipleChoice):
                    attribute = SpecAttributeDefinition(
                        attribute_type=SpecObjectAttributeType.ENUMERATION,
                        identifier=field_identifier,
                        datatype_definition=context.data_types_lookup[
                            field.title
                        ],
                        long_name=field_human_title,
                        multi_valued=True,
                        last_change=context.export_date_str,
                    )
                else:
                    raise NotImplementedError(  # pragma: no cover
                        field
                    ) from None
                attribute_definitions.append(attribute)

            spec_object_type = ReqIFSpecObjectType.create(
                identifier=spec_object_type_identifier,
                long_name=element.tag,
                attribute_definitions=attribute_definitions,
                last_change=context.export_date_str,
            )
            spec_object_types.append(spec_object_type)
            context.map_grammar_node_tags_to_spec_object_type[grammar_document][
                element.tag
            ] = spec_object_type

        assert grammar.parent is not None

        # Using dict as an ordered set.
        spec_relation_tuples: OrderedSet[Tuple[str, Optional[str]]] = (
            OrderedSet()
        )
        for element_ in grammar.elements:
            for relation_ in element_.relations:
                spec_relation_tuples.add(
                    (relation_.relation_type, relation_.relation_role)
                )

        spec_relation_types: List[ReqIFSpecRelationType] = []
        for spec_relation_tuple_ in spec_relation_tuples:
            spec_relation_type_name = (
                spec_relation_tuple_[1]
                if spec_relation_tuple_[1] is not None
                else spec_relation_tuple_[0]
            )
            spec_relation_type = ReqIFSpecRelationType(
                identifier=generate_unique_identifier(spec_relation_type_name),
                last_change=context.export_date_str,
                long_name=spec_relation_type_name,
            )
            spec_relation_types.append(spec_relation_type)
            context.map_spec_relation_tuple_to_spec_relation_type[
                spec_relation_tuple_
            ] = spec_relation_type

        return spec_object_types + spec_relation_types



import typing
from collections import Counter
from dataclasses import dataclass, field
from typing import Dict, List

from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.node import SDocNode
from strictdoc.backend.sdoc.models.section import SDocSection
from strictdoc.helpers.string import create_safe_title_string


@dataclass
class SinglePrefixRequirements:
    requirements_uid_numbers: List[int] = field(default_factory=list)
    requirements_no_uid: List[SDocNode] = field(default_factory=list)


@dataclass
class DocumentStats:
    document: SDocDocument
    requirements_per_prefix: Dict[str, SinglePrefixRequirements] = field(
        default_factory=dict
    )
    sections_without_uid: List[typing.Union[SDocNode, SDocSection]] = field(
        default_factory=list
    )
    section_uids_so_far: typing.Counter[str] = field(default_factory=Counter)


@dataclass
class DocumentTreeStats:
    single_document_stats: List[DocumentStats]

    requirements_per_prefix: Dict[str, SinglePrefixRequirements]
    section_uids_so_far: typing.Counter[str]

    def get_next_requirement_uid(self, prefix: str) -> str:
        next_number = self.get_next_requirement_uid_number(prefix)
        return f"{prefix}{next_number}"

    def get_next_requirement_uid_number(self, prefix: str) -> int:
        if prefix not in self.requirements_per_prefix:
            return 1

        single_prefix_requirements = self.requirements_per_prefix[prefix]

        one_prefix_requirements_uid_numbers = (
            single_prefix_requirements.requirements_uid_numbers
        )
        next_number = (
            max(one_prefix_requirements_uid_numbers) + 1
            if len(one_prefix_requirements_uid_numbers) > 0
            else 1
        )
        return next_number

    def get_auto_section_uid(
        self,
        document_acronym: str,
        section: typing.Union[SDocNode, SDocSection],
    ) -> str:
        if section.reserved_title is None:
            return ""

        section_title = create_safe_title_string(section.reserved_title)
        auto_uid = f"SECTION-{document_acronym}-{section_title}"

        count_so_far = self.section_uids_so_far[auto_uid]
        self.section_uids_so_far[auto_uid] += 1

        if count_so_far >= 1:
            auto_uid += f"-{self.section_uids_so_far[auto_uid]}"
        return auto_uid



"""
@relation(SDOC-SRS-36, scope=file)
"""

# mypy: disable-error-code="operator"
from pathlib import Path
from typing import List, Tuple, Union

from markupsafe import Markup
from pygments import highlight
from pygments.formatters.html import HtmlFormatter
from pygments.lexers import get_lexer_for_filename
from pygments.lexers.c_cpp import CLexer, CppLexer
from pygments.lexers.configs import TOMLLexer
from pygments.lexers.data import YamlLexer
from pygments.lexers.javascript import JavascriptLexer
from pygments.lexers.markup import RstLexer, TexLexer
from pygments.lexers.python import PythonLexer
from pygments.lexers.special import TextLexer
from pygments.lexers.templates import HtmlDjangoLexer
from pygments.util import ClassNotFound

from strictdoc.backend.sdoc_source_code.models.function_range_marker import (
    ForwardFunctionRangeMarker,
    FunctionRangeMarker,
)
from strictdoc.backend.sdoc_source_code.models.line_marker import LineMarker
from strictdoc.backend.sdoc_source_code.models.range_marker import (
    ForwardRangeMarker,
    RangeMarker,
)
from strictdoc.backend.sdoc_source_code.models.source_file_info import (
    SourceFileTraceabilityInfo,
)
from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.source_tree import SourceFile
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.export.html.generators.view_objects.source_file_view_object import (
    SourceFileViewObject,
    SourceLineEntry,
    SourceMarkerTuple,
)
from strictdoc.export.html.html_templates import HTMLTemplates
from strictdoc.export.html.renderers.link_renderer import LinkRenderer
from strictdoc.export.html.renderers.markup_renderer import MarkupRenderer
from strictdoc.helpers.cast import assert_cast
from strictdoc.helpers.timing import measure_performance


class SourceFileViewHTMLGenerator:
    @staticmethod
    def export_to_file(
        *,
        project_config: ProjectConfig,
        source_file: SourceFile,
        traceability_index: TraceabilityIndex,
        html_templates: HTMLTemplates,
    ) -> None:
        if not traceability_index.file_dependency_manager.must_generate(
            source_file.output_file_full_path
        ):
            with measure_performance(
                f"Skip: {source_file.in_doctree_source_file_rel_path}"
            ):
                return

        with measure_performance(
            f"File: {source_file.in_doctree_source_file_rel_path}"
        ):
            document_content = SourceFileViewHTMLGenerator.export(
                project_config=project_config,
                source_file=source_file,
                traceability_index=traceability_index,
                html_templates=html_templates,
            )
            Path(source_file.output_dir_full_path).mkdir(
                parents=True, exist_ok=True
            )
            with open(
                source_file.output_file_full_path, "w", encoding="utf-8"
            ) as file:
                file.write(document_content)

    @staticmethod
    def export(
        *,
        project_config: ProjectConfig,
        source_file: SourceFile,
        traceability_index: TraceabilityIndex,
        html_templates: HTMLTemplates,
    ) -> Markup:
        with open(source_file.full_path, encoding="utf-8") as opened_file:
            source_file_lines = opened_file.readlines()

        pygmented_source_file_lines: List[SourceLineEntry] = []
        pygments_styles: Markup = Markup("")

        trace_info: SourceFileTraceabilityInfo = (
            traceability_index.get_coverage_info(
                source_file.in_doctree_source_file_rel_path_posix
            )
        )

        if len(source_file_lines) > 0:
            (
                pygmented_source_file_lines,
                pygments_styles,
            ) = SourceFileViewHTMLGenerator.get_pygmented_source_lines(
                source_file, source_file_lines, trace_info
            )
        link_renderer = LinkRenderer(
            root_path=source_file.path_depth_prefix,
            static_path=project_config.dir_for_sdoc_assets,
        )
        markup_renderer = MarkupRenderer.create(
            "RST",
            traceability_index,
            link_renderer,
            html_templates,
            project_config,
            None,
        )
        view_object = SourceFileViewObject(
            traceability_index=traceability_index,
            trace_info=trace_info,
            project_config=project_config,
            link_renderer=link_renderer,
            markup_renderer=markup_renderer,
            source_file=source_file,
            pygments_styles=pygments_styles,
            pygmented_source_file_lines=pygmented_source_file_lines,
            jinja_environment=html_templates.jinja_environment(),
        )
        return view_object.render_screen()

    @staticmethod
    def get_pygmented_source_lines(
        source_file: SourceFile,
        source_file_lines: List[str],
        coverage_info: SourceFileTraceabilityInfo,
    ) -> Tuple[
        List[SourceLineEntry],
        Markup,
    ]:
        assert isinstance(source_file, SourceFile)
        assert isinstance(source_file_lines, list)
        assert isinstance(coverage_info, SourceFileTraceabilityInfo)

        if source_file.is_python_file():
            lexer = PythonLexer()
        elif source_file.is_c_file():
            lexer = CLexer()
        elif source_file.is_cpp_file():
            lexer = CppLexer()
        elif source_file.is_tex_file():
            lexer = TexLexer()
        elif source_file.is_toml_file():
            lexer = TOMLLexer()
        elif source_file.is_jinja_file():
            lexer = HtmlDjangoLexer()
        elif source_file.is_javascript_file():
            lexer = JavascriptLexer()
        elif source_file.is_yaml_file():
            lexer = YamlLexer()
        elif source_file.is_rst_file():
            lexer = RstLexer()
        else:
            try:
                lexer = get_lexer_for_filename(source_file.file_name)
            except ClassNotFound:
                lexer = TextLexer()

        # HACK:
        # Otherwise, Pygments will skip the first line as if it does not exist.
        # This behavior surprisingly has an effect on the first line if its empty.
        hack_first_line: bool = False
        if source_file_lines[0] == "\n":
            source_file_lines[0] = " \n"
            hack_first_line = True

        # HACK:
        # Pygments does not process lines if they are empty and are at the end
        # of a file. Adding a marker to the end so that Pygments do not cut the
        # corners.
        source_file_content = "".join(source_file_lines)
        source_file_content_with_marker = source_file_content + "\n###"

        html_formatter = HtmlFormatter()
        pygmented_source_file_content = highlight(
            source_file_content_with_marker, lexer, html_formatter
        )

        # HACK: split content into lines by cutting off the header and footer
        # parts generated by Pygments:
        # <div class="highlight"><pre> and </pre></div>
        # TODO: Implement proper splitting.
        start_pattern = '<div class="highlight"><pre>'
        end_pattern = "</pre></div>\n"
        assert pygmented_source_file_content.startswith(start_pattern)
        assert pygmented_source_file_content.endswith(end_pattern), (
            f"{pygmented_source_file_content}"
        )

        slice_start = len(start_pattern)
        slice_end = len(pygmented_source_file_content) - len(end_pattern)
        pygmented_source_file_content = pygmented_source_file_content[
            slice_start:slice_end
        ]
        pygmented_source_file_lines: List[Union[str, SourceMarkerTuple]] = list(
            pygmented_source_file_content.split("\n")
        )
        if hack_first_line:
            pygmented_source_file_lines[0] = "<span></span>"

        if pygmented_source_file_lines[-1] == "":
            pygmented_source_file_lines.pop()
        assert "###" in pygmented_source_file_lines[-1], (
            "Expected marker to be in place."
        )
        # Pop ###, pop "\n"
        pygmented_source_file_lines.pop()
        if pygmented_source_file_lines[-1] == "":
            pygmented_source_file_lines.pop()

        assert len(pygmented_source_file_lines) == len(source_file_lines), (
            f"Something went wrong when running Pygments against "
            f"the source file: "
            f"{len(pygmented_source_file_lines)} == {len(source_file_lines)}, "
            f"{pygmented_source_file_lines} == {source_file_lines}."
        )

        for marker in coverage_info.markers:
            marker_line = (
                marker.ng_range_line_begin
                if marker.is_begin()
                else marker.ng_range_line_end
            )
            assert isinstance(marker_line, int)
            source_marker_tuple: SourceMarkerTuple
            if isinstance(
                pygmented_source_file_lines[marker_line - 1], SourceMarkerTuple
            ):
                source_marker_tuple = assert_cast(
                    pygmented_source_file_lines[marker_line - 1],
                    SourceMarkerTuple,
                )
            else:
                pygmented_source_file_line = assert_cast(
                    pygmented_source_file_lines[marker_line - 1], str
                )
                assert marker.ng_range_line_begin is not None
                assert marker.ng_range_line_end is not None
                source_marker_tuple = SourceMarkerTuple(
                    ng_range_line_begin=marker.ng_range_line_begin,
                    ng_range_line_end=marker.ng_range_line_end,
                    source_line=Markup(pygmented_source_file_line),
                    markers=[],
                )
                pygmented_source_file_lines[marker_line - 1] = (
                    source_marker_tuple
                )

            if isinstance(
                marker,
                (
                    ForwardRangeMarker,
                    ForwardFunctionRangeMarker,
                    FunctionRangeMarker,
                    RangeMarker,
                    LineMarker,
                ),
            ):
                source_marker_tuple.markers.append(marker)
                continue

            raise NotImplementedError(marker)  # pragma: no cover

        pygments_styles = (
            f"/* Lexer: {lexer.name} */\n"
            + html_formatter.get_style_defs(".highlight")
        )

        return [
            SourceFileViewHTMLGenerator.mark_safe(line)
            for line in pygmented_source_file_lines
        ], Markup(pygments_styles)

    @staticmethod
    def mark_safe(
        line: Union[str, SourceMarkerTuple],
    ) -> Union[Markup, SourceMarkerTuple]:
        return line if isinstance(line, SourceMarkerTuple) else Markup(line)



"""
@relation(SDOC-SRS-33, scope=file)
"""

from typing import Any, Dict, List, Optional

from strictdoc.backend.sdoc_source_code.models.range_marker import RangeMarker
from strictdoc.helpers.file_stats import SourceFileStats


class ParseContext:
    def __init__(
        self, filename: Optional[str], file_stats: SourceFileStats
    ) -> None:
        self.filename: Optional[str] = filename
        self.file_stats: SourceFileStats = file_stats
        self.markers: List[Any] = []
        self.marker_stack: List[RangeMarker] = []
        self.map_reqs_to_markers: Dict[str, Any] = {}



from typing import List, Optional, Set

from markupsafe import Markup
from starlette.datastructures import FormData

from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.document_grammar import (
    DocumentGrammar,
)
from strictdoc.backend.sdoc.models.grammar_element import GrammarElement
from strictdoc.core.project_config import ProjectConfig
from strictdoc.export.html.form_objects.rows.row_with_grammar_element_form_object import (
    RowWithGrammarElementFormObject,
)
from strictdoc.export.html.html_templates import JinjaEnvironment
from strictdoc.helpers.auto_described import auto_described
from strictdoc.helpers.cast import assert_cast
from strictdoc.helpers.form_data import (
    ParsedFormData,
    parse_form_data,
)
from strictdoc.helpers.mid import MID
from strictdoc.helpers.string import is_uppercase_underscore_string
from strictdoc.server.error_object import ErrorObject
from strictdoc.server.helpers.turbo import render_turbo_stream


@auto_described
class GrammarElementFormField:
    def __init__(
        self,
        is_new: bool,
        field_mid: str,
        field_name: str,
        document_mid: str,
    ):
        self.is_new: bool = is_new
        self.field_mid: str = field_mid
        self.field_name: str = field_name
        self.document_mid: str = document_mid

    @staticmethod
    def create(
        *,
        is_new: bool,
        grammar_element: GrammarElement,
        document_mid: str,
    ) -> "GrammarElementFormField":
        return GrammarElementFormField(
            is_new=is_new,
            field_mid=grammar_element.mid,
            field_name=grammar_element.tag,
            document_mid=document_mid,
        )

    def get_input_field_name(self) -> str:
        return f"document_grammar_element_field[{self.field_mid}][field_name]"

    def get_input_field_is_new(self) -> str:
        return f"document_grammar_element_field[{self.field_mid}][is_new]"

    def get_is_new_as_string(self) -> str:
        return "true" if self.is_new else "false"


@auto_described
class GrammarFormObject(ErrorObject):
    def __init__(
        self,
        *,
        document_mid: str,
        fields: List[GrammarElementFormField],
        project_config: ProjectConfig,
        jinja_environment: JinjaEnvironment,
        imported_grammar_file: Optional[str],
    ):
        assert isinstance(document_mid, str), document_mid
        super().__init__()
        self.document_mid = document_mid
        self.fields: List[GrammarElementFormField] = fields
        self.project_config: ProjectConfig = project_config
        self.jinja_environment: JinjaEnvironment = jinja_environment
        self.imported_grammar_file: Optional[str] = imported_grammar_file

    @staticmethod
    def create_from_request(
        *,
        document_mid: str,
        request_form_data: FormData,
        project_config: ProjectConfig,
        jinja_environment: JinjaEnvironment,
    ) -> "GrammarFormObject":
        form_object_fields: List[GrammarElementFormField] = []
        request_form_data_as_list = [
            (field_name, field_value)
            for field_name, field_value in request_form_data.multi_items()
        ]
        request_form_dict: ParsedFormData = assert_cast(
            parse_form_data(request_form_data_as_list), dict
        )

        document_grammar_fields = request_form_dict[
            "document_grammar_element_field"
        ]

        if not isinstance(document_grammar_fields, dict):
            raise TypeError(
                f"Expected a dict, but got {type(document_grammar_fields)}"
            )

        for field_mid, field_dict in document_grammar_fields.items():
            assert isinstance(field_dict, dict), type(field_dict)

            is_new = field_dict["is_new"] == "true"
            field_name = assert_cast(field_dict["field_name"], str)

            form_object_field = GrammarElementFormField(
                is_new=is_new,
                field_mid=field_mid,
                field_name=field_name,
                document_mid=document_mid,
            )
            form_object_fields.append(form_object_field)

        form_object = GrammarFormObject(
            document_mid=document_mid,
            fields=form_object_fields,
            project_config=project_config,
            jinja_environment=jinja_environment,
            imported_grammar_file=None,
        )
        return form_object

    @staticmethod
    def create_from_document(
        *,
        document: SDocDocument,
        project_config: ProjectConfig,
        jinja_environment: JinjaEnvironment,
    ) -> "GrammarFormObject":
        assert isinstance(document, SDocDocument)
        assert isinstance(document.grammar, DocumentGrammar)

        grammar: DocumentGrammar = document.grammar

        grammar_element_form_fields: List[GrammarElementFormField] = []

        for element_ in grammar.elements:
            grammar_form_field = GrammarElementFormField.create(
                is_new=False,
                grammar_element=element_,
                document_mid=document.reserved_mid,
            )
            grammar_element_form_fields.append(grammar_form_field)

        return GrammarFormObject(
            document_mid=document.reserved_mid,
            fields=grammar_element_form_fields,
            project_config=project_config,
            jinja_environment=jinja_environment,
            imported_grammar_file=grammar.import_from_file,
        )

    def validate(self) -> bool:
        fields_so_far: Set[str] = set()
        for field in self.fields:
            if len(field.field_name) == 0:
                self.add_error(
                    field.get_input_field_name(),
                    "Provide a name for the grammar element.",
                )
                continue

            if not is_uppercase_underscore_string(field.field_name):
                self.add_error(
                    field.get_input_field_name(),
                    (
                        "Grammar element title shall consist of "
                        "uppercase letters, digits and single underscores."
                    ),
                )

            if field.field_name in fields_so_far:
                self.add_error(
                    field.get_input_field_name(),
                    f"Grammar element {field.field_name} is not unique.",
                )
            else:
                fields_so_far.add(field.field_name)

        return len(self.errors) == 0

    def render(self) -> Markup:
        rendered_template = self.jinja_environment.render_template_as_markup(
            "components/grammar_form/index.jinja", form_object=self
        )
        return render_turbo_stream(
            content=rendered_template, action="update", target="modal"
        )

    def render_row_with_grammar_element(
        self, field: GrammarElementFormField
    ) -> Markup:
        form_object = RowWithGrammarElementFormObject(
            field=field,
            errors=self.errors,
            jinja_environment=self.jinja_environment,
        )
        return form_object.render()

    def render_row_with_new_grammar_element(self) -> str:
        field: GrammarElementFormField = GrammarElementFormField(
            is_new=True,
            field_mid=MID.create(),
            field_name="",
            document_mid=self.document_mid,
        )
        form_object = RowWithGrammarElementFormObject(
            field=field,
            errors=self.errors,
            jinja_environment=self.jinja_environment,
        )
        rendered_template: str = form_object.render()
        return render_turbo_stream(
            content=rendered_template,
            action="append",
            target="document__editable_grammar_elements",
        )

    @staticmethod
    def render_close_form() -> str:
        return render_turbo_stream(
            content="",
            action="update",
            target="modal",
        )



from typing import List


def convert_function_name_to_gtest_macro(input_str: str) -> List[str]:
    """
    Convert a Google Test name from the JUnit XML format to a C++ macro definition.

    Converts a Google Test-produced test name from JUnit XML as follows:
    TestPrtMath.TransitionDistance -> TEST_F(TestPrtMath, TransitionDistance)
    The TEST_F... format is how tree-sitter-cpp parses this macro from C++ code,
    and this is the format used for matching the test function with the
    corresponding test result.
    """
    assert isinstance(input_str, str)
    assert len(input_str) > 0

    # Google Test parts show up like this in the XML:
    # XML: MyTestPattern/MyTestHelperPattern.TestName3/3  # noqa: ERA001
    # C++: TEST_P(MyTestHelperPattern, TestName3)
    is_pattern_function = "/" in input_str
    if is_pattern_function:
        input_str_parts = input_str.split("/")
        input_str = input_str_parts[1]

    if "." not in input_str:
        raise ValueError(f"Input string must contain a dot: '{input_str}'.")

    parts = input_str.split(".")
    if len(parts) != 2:
        raise ValueError(
            f"Input string must contain exactly one dot: '{input_str}'."
        )

    if is_pattern_function:
        return [
            f"TEST_P({parts[0]}, {parts[1]})",
        ]

    return [f"TEST({parts[0]}, {parts[1]})", f"TEST_F({parts[0]}, {parts[1]})"]



"""
@relation(SDOC-SRS-72, scope=file)
"""


class ReqIFProfile:
    P01_SDOC = "p01_sdoc"

    ALL = [P01_SDOC]


SDOC_SPEC_OBJECT_TYPE_SINGLETON = "REQUIREMENT_OR_SECTION"


class SDocRequirementReservedField:
    UID = "UID"
    TITLE = "TITLE"
    STATEMENT = "STATEMENT"
    COMMENT = "COMMENT"
    CREATED_BY = "CREATED_BY"


class ReqIFReservedField:
    """
    Captures some of the most important field naming conventions by ReqIF.
    """

    UID = "ReqIF.ForeignID"

    # The chapter name appears with elements that represent chapters/sections.
    # The name usually appears with non-section elements.
    CHAPTER_NAME = "ReqIF.ChapterName"
    NAME = "ReqIF.Name"

    TEXT = "ReqIF.Text"

    CREATED_BY = "ReqIF.ForeignCreatedBy"
    COMMENT_NOTES = "NOTES"


SDOC_TO_REQIF_FIELD_MAP = {
    SDocRequirementReservedField.UID: ReqIFReservedField.UID,
    SDocRequirementReservedField.STATEMENT: ReqIFReservedField.TEXT,
    SDocRequirementReservedField.COMMENT: ReqIFReservedField.COMMENT_NOTES,
    SDocRequirementReservedField.CREATED_BY: ReqIFReservedField.CREATED_BY,
}

REQIF_MAP_TO_SDOC_FIELD_MAP = {
    ReqIFReservedField.UID: SDocRequirementReservedField.UID,
    ReqIFReservedField.NAME: SDocRequirementReservedField.TITLE,
    ReqIFReservedField.CHAPTER_NAME: SDocRequirementReservedField.TITLE,
    ReqIFReservedField.TEXT: SDocRequirementReservedField.STATEMENT,
    ReqIFReservedField.COMMENT_NOTES: SDocRequirementReservedField.COMMENT,
    ReqIFReservedField.CREATED_BY: SDocRequirementReservedField.CREATED_BY,
}

DEFAULT_SDOC_GRAMMAR_FIELDS = [
    "ReqIF.ForeignID",
    "LEVEL",
    "STATUS",
    "TAGS",
    "ReqIF.Name",
    "ReqIF.Text",
    "RATIONALE",
    "NOTES",
    "ReqIF.ChapterName",
]

SDOC_SPECIFICATION_TYPE_SINGLETON = "SDOC_SPECIFICATION_TYPE_SINGLETON"


def map_sdoc_field_title_to_reqif_field_title(
    sdoc_field_title: str, is_composite: bool
) -> str:
    """
    Map how SDoc field titles are converted to ReqIF field titles.
    """

    # Assumption: The composite types will most of the type be chapters/sections,
    # so mapping them to ReqIF Chapter Name.
    # If a user requires a more distinct mapping, we would need to introduce
    # a configuration parameter with an explicit map.
    if sdoc_field_title == "TITLE":
        if is_composite:
            return ReqIFReservedField.CHAPTER_NAME
        return ReqIFReservedField.NAME
    return SDOC_TO_REQIF_FIELD_MAP.get(sdoc_field_title) or sdoc_field_title


def map_reqif_field_title_to_sdoc_field_title(reqif_field_title: str) -> str:
    return (
        REQIF_MAP_TO_SDOC_FIELD_MAP.get(reqif_field_title) or reqif_field_title
    )



"""
@relation(SDOC-SRS-55, scope=file)
"""

import datetime
from copy import copy
from dataclasses import dataclass
from typing import Dict, List, Optional, Set, Tuple, Union

from textx import TextXSyntaxError

from strictdoc.backend.sdoc.document_reference import DocumentReference
from strictdoc.backend.sdoc.error_handling import get_textx_syntax_error_message
from strictdoc.backend.sdoc.free_text_reader import SDFreeTextReader
from strictdoc.backend.sdoc.models.anchor import Anchor
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.free_text import FreeTextContainer
from strictdoc.backend.sdoc.models.grammar_element import GrammarElement
from strictdoc.backend.sdoc.models.inline_link import InlineLink
from strictdoc.backend.sdoc.models.model import SDocDocumentIF, SDocSectionIF
from strictdoc.backend.sdoc.models.node import SDocNode, SDocNodeField
from strictdoc.backend.sdoc.models.object_factory import SDocObjectFactory
from strictdoc.backend.sdoc.models.reference import (
    Reference,
)
from strictdoc.backend.sdoc.models.section import SDocSection
from strictdoc.core.constants import (
    GraphLinkType,
)
from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.traceability_index import (
    TraceabilityIndex,
)
from strictdoc.core.transforms.constants import NodeCreationOrder
from strictdoc.core.transforms.validation_error import (
    SingleValidationError,
)
from strictdoc.export.html.form_objects.requirement_form_object import (
    RequirementFormField,
    RequirementFormFieldType,
    RequirementFormObject,
)
from strictdoc.export.rst.rst_to_html_fragment_writer import (
    RstToHtmlFragmentWriter,
)
from strictdoc.helpers.cast import assert_cast
from strictdoc.helpers.file_modification_time import set_file_modification_time
from strictdoc.helpers.mid import MID


@dataclass
class CreateNodeInfo:
    whereto: str
    requirement_mid: str
    reference_mid: str


@dataclass
class UpdateNodeInfo:
    node_to_update: SDocNode


class UpdateRequirementActionObject:
    def __init__(self) -> None:
        self.existing_references_uids: Set[Tuple[str, str, Optional[str]]] = (
            set()
        )
        self.reference_ids_to_remove: Set[Tuple[str, str, Optional[str]]] = (
            set()
        )
        self.removed_uid_parent_documents_to_update: Set[SDocDocument] = set()
        # All requirements that have to be updated. This set includes
        # the requirement itself, all links it was linking to
        # (for deleted links) and all links it is linking to now
        # (including new links).
        self.this_document_requirements_to_update: Set[SDocNode] = set()


@dataclass
class CreateOrUpdateNodeResult:
    this_document_requirements_to_update: List[SDocNode]


class CreateOrUpdateNodeCommand:
    def __init__(
        self,
        *,
        form_object: RequirementFormObject,
        node_info: Union[CreateNodeInfo, UpdateNodeInfo],
        context_document: SDocDocument,
        traceability_index: TraceabilityIndex,
        project_config: ProjectConfig,
    ) -> None:
        self.form_object: RequirementFormObject = form_object
        self.node_info: Union[CreateNodeInfo, UpdateNodeInfo] = node_info
        self.context_document: SDocDocument = context_document
        self.traceability_index: TraceabilityIndex = traceability_index
        self.project_config: ProjectConfig = project_config

    def perform(self) -> Optional[CreateOrUpdateNodeResult]:
        form_object: RequirementFormObject = self.form_object
        node_to_update_or_none: Optional[SDocNode] = (
            self.node_info.node_to_update
            if isinstance(self.node_info, UpdateNodeInfo)
            else None
        )

        traceability_index: TraceabilityIndex = self.traceability_index

        map_form_to_requirement_fields: Dict[
            RequirementFormField, Optional[FreeTextContainer]
        ] = {}
        this_node_unique_anchors: Set[str] = set()
        for field_bucket_ in form_object.fields.values():
            field_: RequirementFormField
            for field_ in field_bucket_:
                if field_.field_type != RequirementFormFieldType.MULTILINE:
                    continue
                (
                    parsed_html,
                    rst_error,
                ) = RstToHtmlFragmentWriter(
                    project_config=self.project_config,
                    context_document=self.context_document,
                ).write_with_validation(field_.field_value)
                if parsed_html is None:
                    assert rst_error is not None
                    form_object.add_error(field_.field_name, rst_error)
                else:
                    try:
                        free_text_container: Optional[FreeTextContainer] = (
                            SDFreeTextReader.read(field_.field_value)
                            if len(field_.field_value) > 0
                            else None
                        )
                        map_form_to_requirement_fields[field_] = (
                            free_text_container
                        )
                        if free_text_container is None:
                            continue
                        anchors: List[Anchor] = []
                        for part in free_text_container.parts:
                            if isinstance(part, InlineLink):
                                linked_to_node = traceability_index.get_linkable_node_by_uid_weak(
                                    part.link
                                )
                                if linked_to_node is None:
                                    form_object.add_error(
                                        field_.field_name,
                                        "A LINK points to a node that does "
                                        f"not exist: '{part.link}'.",
                                    )
                            elif isinstance(part, Anchor):
                                if part.value in this_node_unique_anchors:
                                    form_object.add_error(
                                        field_.field_name,
                                        "The node fields contain duplicate anchor: "
                                        f"'{part.value}'.",
                                    )
                                    break
                                this_node_unique_anchors.add(part.value)
                                anchors.append(part)
                            else:
                                pass
                        if anchors is not None:
                            try:
                                traceability_index.validate_node_against_anchors(
                                    node=node_to_update_or_none,
                                    new_anchors=anchors,
                                )
                            except (
                                SingleValidationError
                            ) as anchors_validation_error:
                                form_object.add_error(
                                    field_.field_name,
                                    anchors_validation_error.args[0],
                                )
                    except TextXSyntaxError as exception:
                        form_object.add_error(
                            field_.field_name,
                            get_textx_syntax_error_message(exception),
                        )

        if form_object.any_errors():
            return None

        requirement: SDocNode
        document: SDocDocument
        existing_uid: Optional[str]
        existing_node_fields: List[SDocNodeField] = []
        if isinstance(self.node_info, UpdateNodeInfo):
            requirement = self.node_info.node_to_update
            document = assert_cast(requirement.get_document(), SDocDocument)

            existing_uid = requirement.reserved_uid

            existing_node_fields = list(requirement.enumerate_fields())

            # Clearing all existing fields because they will be recreated from
            # scratch from the form data.
            requirement.ordered_fields_lookup.clear()

            self.populate_node_fields_from_form_object(
                requirement, form_object, map_form_to_requirement_fields
            )

            # Updating Traceability Index: UID.
            traceability_index.update_requirement_uid(
                requirement=requirement, old_uid=existing_uid
            )
        else:
            reference_node: Union[SDocDocument, SDocSection] = (
                traceability_index.get_node_by_mid(
                    MID(self.node_info.reference_mid)
                )
            )
            document = traceability_index.get_node_by_mid(
                MID(form_object.document_mid)
            )
            parent: Union[SDocDocumentIF, SDocSectionIF]
            if self.node_info.whereto == NodeCreationOrder.CHILD:
                parent = reference_node
                insert_to_idx = len(parent.section_contents)
            elif self.node_info.whereto == NodeCreationOrder.BEFORE:
                # Be aware of an edge case besides all normal cases:
                # A reference node can be a root node of an included document.
                if not isinstance(reference_node, SDocDocument):
                    parent = reference_node.parent
                    insert_to_idx = parent.section_contents.index(
                        reference_node
                    )
                else:
                    assert (
                        reference_node.ng_including_document_from_file
                        is not None
                    )
                    parent = (
                        reference_node.ng_including_document_from_file.parent
                    )
                    assert (
                        reference_node.ng_including_document_from_file
                        is not None
                    )
                    insert_to_idx = parent.section_contents.index(
                        reference_node.ng_including_document_from_file
                    )
            elif self.node_info.whereto == NodeCreationOrder.AFTER:
                if isinstance(reference_node, SDocDocument):
                    assert reference_node.document_is_included()
                    assert reference_node.ng_including_document_from_file
                    parent = (
                        reference_node.ng_including_document_from_file.parent
                    )
                    insert_to_idx = (
                        parent.section_contents.index(
                            reference_node.ng_including_document_from_file
                        )
                        + 1
                    )
                else:
                    parent = reference_node.parent
                    insert_to_idx = (
                        parent.section_contents.index(reference_node) + 1
                    )
            else:
                raise NotImplementedError

            # Reset the 'needs generation' flag on all documents.
            for document_ in traceability_index.document_tree.document_list:
                assert document_.meta is not None
                set_file_modification_time(
                    document_.meta.input_doc_full_path,
                    datetime.datetime.today(),
                )

            # FIXME: It is better to have a general create_node method because
            #        we are dealing with arbitrary nodes, not only Requirement.
            requirement = SDocObjectFactory.create_requirement(
                parent=parent, node_type=form_object.element_type
            )
            requirement.reserved_mid = MID(form_object.requirement_mid)
            if document.config.enable_mid:
                requirement.mid_permanent = True

            assert document.grammar is not None
            grammar_element: GrammarElement = document.grammar.elements_by_type[
                form_object.element_type
            ]

            requirement.is_composite = (
                grammar_element.property_is_composite is True
            )
            requirement.ng_document_reference = DocumentReference()
            requirement.ng_document_reference.set_document(document)
            requirement.ng_including_document_reference = (
                document.ng_including_document_reference
            )

            parent.section_contents.insert(insert_to_idx, requirement)

            self.populate_node_fields_from_form_object(
                requirement, form_object, map_form_to_requirement_fields
            )
            traceability_index.create_requirement(requirement=requirement)

        action_object = UpdateRequirementActionObject()
        action_object.existing_references_uids.update(
            requirement.get_requirement_reference_uids()
        )
        action_object.reference_ids_to_remove = copy(
            action_object.existing_references_uids
        )
        action_object.this_document_requirements_to_update = {requirement}

        references: List[Reference] = form_object.get_requirement_relations(
            requirement
        )
        if len(references) > 0:
            requirement.relations = references
        else:
            requirement.relations = []

        for document_ in traceability_index.document_tree.document_list:
            assert document_.meta is not None
            set_file_modification_time(
                document_.meta.input_doc_full_path, datetime.datetime.today()
            )

        # Updating Traceability Index: Links.
        for reference_field in form_object.reference_fields:
            ref_uid = reference_field.field_value
            ref_role: Optional[str] = (
                reference_field.field_role
                if reference_field.field_role is not None
                and len(reference_field.field_role) > 0
                else None
            )
            # If a link is in the form, we don't want to remove it.
            if (
                reference_field.field_type,
                ref_uid,
                ref_role,
            ) in action_object.reference_ids_to_remove:
                action_object.reference_ids_to_remove.remove(
                    (reference_field.field_type, ref_uid, ref_role)
                )
            # If a link is already in the requirement and traceability index,
            # there is nothing to do.
            if (
                reference_field.field_type,
                ref_uid,
                ref_role,
            ) in action_object.existing_references_uids:
                continue
            if reference_field.field_type == "Parent":
                traceability_index.update_requirement_parent_uid(
                    requirement=requirement,
                    parent_uid=ref_uid,
                    role=reference_field.field_role
                    if len(reference_field.field_role) > 0
                    else None,
                )
            elif reference_field.field_type == "Child":
                traceability_index.update_requirement_child_uid(
                    requirement=requirement,
                    child_uid=ref_uid,
                    role=reference_field.field_role
                    if len(reference_field.field_role) > 0
                    else None,
                )
            elif reference_field.field_type == "File":
                pass
            else:
                raise AssertionError(f"Must not reach here: {reference_field}")

        # Calculate which documents and requirements have to be regenerated.
        for (
            _,
            reference_id_to_remove,
            _,
        ) in action_object.reference_ids_to_remove:
            removed_uid_parent_requirement: SDocNode = (
                traceability_index.graph_database.get_link_value(
                    link_type=GraphLinkType.UID_TO_NODE,
                    lhs_node=reference_id_to_remove,
                )
            )
            removed_uid_parent_requirement_document = assert_cast(
                removed_uid_parent_requirement.get_document(), SDocDocument
            )
            action_object.removed_uid_parent_documents_to_update.add(
                removed_uid_parent_requirement_document
            )
            # If a link was pointing towards a parent requirement in this
            # document, we will have to re-render it now.
            if removed_uid_parent_requirement_document == document:
                action_object.this_document_requirements_to_update.add(
                    removed_uid_parent_requirement
                )

        for (
            relation_type_,
            reference_id_to_remove,
            reference_id_to_remove_role,
        ) in action_object.reference_ids_to_remove:
            if relation_type_ == "Parent":
                traceability_index.remove_requirement_parent_uid(
                    requirement=requirement,
                    parent_uid=reference_id_to_remove,
                    role=reference_id_to_remove_role,
                )
            elif relation_type_ == "Child":
                traceability_index.remove_requirement_child_uid(
                    requirement=requirement,
                    child_uid=reference_id_to_remove,
                    role=reference_id_to_remove_role,
                )

        # Rendering back the Turbo template for each changed requirement.
        for reference_field in form_object.reference_fields:
            if reference_field.field_type not in ("Parent", "Child"):
                continue
            ref_uid = reference_field.field_value

            node: SDocNode = traceability_index.graph_database.get_link_value(
                link_type=GraphLinkType.UID_TO_NODE,
                lhs_node=ref_uid,
            )

            if node.get_document() == document:
                action_object.this_document_requirements_to_update.add(node)

        self._update_traceability_index_with_links_and_anchors(
            requirement, existing_node_fields
        )

        traceability_index.update_last_updated()

        return CreateOrUpdateNodeResult(
            this_document_requirements_to_update=list(
                action_object.this_document_requirements_to_update
            )
        )

    def _update_traceability_index_with_links_and_anchors(
        self, updated_node: SDocNode, existing_node_fields: List[SDocNodeField]
    ) -> None:
        traceability_index = self.traceability_index

        existing_anchor_uids: Set[str] = set()
        existing_links: List[InlineLink] = []
        for node_field_ in existing_node_fields:
            for part_ in node_field_.parts:
                if isinstance(part_, Anchor):
                    existing_anchor_uids.add(part_.value)
                elif isinstance(part_, InlineLink):
                    existing_links.append(part_)

        for existing_link_ in existing_links:
            traceability_index.remove_inline_link(existing_link_)

        for existing_anchor_uid in existing_anchor_uids:
            traceability_index.remove_anchor_by_uid(existing_anchor_uid)

        for node_field_ in updated_node.enumerate_fields():
            for part_ in node_field_.parts:
                if isinstance(part_, Anchor):
                    # Since this is a new section, we just need to register the
                    # new anchor. By this time, we know that there is no
                    # existing anchor with this name.
                    traceability_index.update_with_anchor(part_)
                elif isinstance(part_, InlineLink):
                    traceability_index.create_inline_link(part_)

    def populate_node_fields_from_form_object(
        self,
        node: SDocNode,
        form_object: RequirementFormObject,
        map_form_to_requirement_fields: Dict[
            RequirementFormField, Optional[FreeTextContainer]
        ],
    ) -> None:
        # FIXME: Leave only one method based on set_field_value().
        for form_field_name, form_fields in form_object.fields.items():
            for form_field_index, form_field in enumerate(form_fields):
                if form_field.field_type != RequirementFormFieldType.MULTILINE:
                    node.set_field_value(
                        field_name=form_field_name,
                        form_field_index=form_field_index,
                        value=form_field.field_value,
                    )
                    continue

                free_text_content: Optional[FreeTextContainer] = (
                    map_form_to_requirement_fields[form_field]
                )
                requirement_field: Optional[SDocNodeField] = (
                    SDocNodeField(
                        node,
                        field_name=form_field_name,
                        parts=free_text_content.parts,
                        multiline__="true"
                        if form_field.field_type
                        == RequirementFormFieldType.MULTILINE
                        else None,
                    )
                    if free_text_content is not None
                    else None
                )
                node.set_field_value(
                    field_name=form_field_name,
                    form_field_index=form_field_index,
                    value=requirement_field,
                )
                if (
                    requirement_field is not None
                    and free_text_content is not None
                ):
                    for part_ in requirement_field.parts:
                        if isinstance(part_, str):
                            continue
                        part_.parent = requirement_field



import datetime
import os


def get_file_modification_time(path: str) -> datetime.datetime:
    timestamp: float = os.path.getmtime(path)
    time = datetime.datetime.fromtimestamp(timestamp)
    return time


def set_file_modification_time(path: str, mod_time: datetime.datetime) -> None:
    assert os.path.isfile(path), path

    # Convert datetime to a timestamp.
    timestamp = mod_time.timestamp()
    # Set the modification time (and optionally access time)
    os.utime(path, times=(timestamp, timestamp))



"""
@relation(SDOC-SRS-141, scope=file)
"""

from dataclasses import dataclass, field
from typing import Dict, List, Optional, Union

from strictdoc.backend.sdoc_source_code.models.function_range_marker import (
    FunctionRangeMarker,
)
from strictdoc.backend.sdoc_source_code.models.line_marker import LineMarker
from strictdoc.backend.sdoc_source_code.models.range_marker import (
    RangeMarker,
)


@dataclass
class SourceNode:
    entity_name: Optional[str]
    markers: List[Union[FunctionRangeMarker, RangeMarker, LineMarker]] = field(
        default_factory=list
    )
    fields: Dict[str, str] = field(default_factory=dict)



from typing import Dict, List

from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.document_grammar import (
    DocumentGrammar,
)
from strictdoc.backend.sdoc.models.grammar_element import GrammarElement
from strictdoc.core.traceability_index import (
    TraceabilityIndex,
)
from strictdoc.export.html.form_objects.grammar_form_object import (
    GrammarFormObject,
)


class UpdateGrammarCommand:
    def __init__(
        self,
        *,
        form_object: GrammarFormObject,
        document: SDocDocument,
        traceability_index: TraceabilityIndex,
    ):
        self.form_object: GrammarFormObject = form_object
        self.document: SDocDocument = document
        self.traceability_index: TraceabilityIndex = traceability_index

    def perform(self) -> bool:
        form_object: GrammarFormObject = self.form_object
        document: SDocDocument = self.document

        form_element_names = map(
            lambda field_: field_.field_name, form_object.fields
        )

        map_existing_elements_by_name: Dict[str, GrammarElement] = {}
        assert document.grammar is not None
        for grammar_element_ in document.grammar.elements:
            map_existing_elements_by_name[grammar_element_.tag] = (
                grammar_element_
            )

        updated_grammar_elements: List[GrammarElement] = []
        for form_element_name_ in form_element_names:
            if form_element_name_ in map_existing_elements_by_name:
                updated_grammar_elements.append(
                    map_existing_elements_by_name[form_element_name_]
                )
            else:
                updated_grammar_elements.append(
                    GrammarElement.create_default(form_element_name_)
                )

        new_grammar = DocumentGrammar(
            parent=document, elements=updated_grammar_elements
        )
        document.grammar = new_grammar

        return True



import argparse
import os
from typing import Any, List, Optional

from strictdoc.cli.command_parser_builder import (
    CommandParserBuilder,
    SDocArgumentParser,
)
from strictdoc.helpers.auto_described import auto_described
from strictdoc.helpers.cast import assert_cast
from strictdoc.helpers.net import is_valid_host


class CLIValidationError(Exception):
    pass


class ImportReqIFCommandConfig:
    def __init__(
        self,
        input_path: str,
        output_path: str,
        profile: str,
        reqif_enable_mid: bool,
        reqif_import_markup: Optional[str],
    ):
        self.input_path: str = input_path
        self.output_path: str = output_path
        self.profile: Optional[str] = profile
        self.reqif_enable_mid: bool = reqif_enable_mid
        self.reqif_import_markup: Optional[str] = reqif_import_markup


class ManageAutoUIDCommandConfig:
    def __init__(
        self,
        *,
        input_path: str,
        config_path: Optional[str],
        include_sections: bool,
    ):
        self.input_path: str = input_path
        self._config_path: Optional[str] = config_path
        self.include_sections: bool = include_sections

    def get_path_to_config(self) -> str:
        path_to_input_dir: str = self.input_path
        if os.path.isfile(path_to_input_dir):
            path_to_input_dir = os.path.dirname(path_to_input_dir)
        path_to_config = (
            self._config_path
            if self._config_path is not None
            else path_to_input_dir
        )
        return path_to_config

    def validate(self) -> None:
        if self._config_path is not None and not os.path.exists(
            self._config_path
        ):
            raise CLIValidationError(
                "Provided path to a configuration file does not exist: "
                f"{self._config_path}"
            )


class ImportExcelCommandConfig:
    def __init__(
        self, input_path: str, output_path: str, parser: SDocArgumentParser
    ) -> None:
        self.input_path: str = input_path
        self.output_path: str = output_path
        self.parser: SDocArgumentParser = parser


@auto_described
class ServerCommandConfig:
    def __init__(
        self,
        *,
        input_path: str,
        output_path: Optional[str],
        config_path: Optional[str],
        reload: bool,
        host: Optional[str],
        port: Optional[int],
    ):
        self._input_path: str = input_path
        self.output_path: Optional[str] = output_path
        self._config_path: Optional[str] = config_path
        self.reload: bool = reload
        self.host: Optional[str] = host
        self.port: Optional[int] = port

    def get_full_input_path(self) -> str:
        return os.path.abspath(self._input_path)

    def get_path_to_config(self) -> str:
        return (
            self._config_path
            if self._config_path is not None
            else self._input_path
        )

    def validate(self) -> None:
        if not os.path.exists(self._input_path):
            raise CLIValidationError(
                f"Provided input path does not exist: {self._input_path}"
            )

        if self._config_path is not None and not os.path.exists(
            self._config_path
        ):
            raise CLIValidationError(
                "Provided path to a configuration file does not exist: "
                f"{self._config_path}"
            )

        if (host_ := self.host) is not None:
            if not is_valid_host(host_):
                raise CLIValidationError(
                    f"Provided 'host' argument is not a valid host: {host_}"
                )


@auto_described
class ExportCommandConfig:
    def __init__(
        self,
        input_paths: List[str],
        output_dir: Optional[str],
        config_path: Optional[str],
        project_title: Optional[str],
        formats: List[str],
        fields: List[str],
        generate_bundle_document: bool,
        no_parallelization: bool,
        enable_mathjax: bool,
        included_documents: bool,
        filter_requirements: Optional[str],
        filter_sections: Optional[str],
        reqif_profile: Optional[str],
        reqif_multiline_is_xhtml: bool,
        reqif_enable_mid: bool,
        view: Optional[str],
        chromedriver: Optional[str],
    ):
        assert isinstance(input_paths, list), f"{input_paths}"
        self.input_paths: List[str] = input_paths
        self.output_dir: Optional[str] = output_dir
        self._config_path: Optional[str] = config_path
        self.project_title: Optional[str] = project_title
        self.formats: List[str] = formats
        self.fields: List[str] = fields
        self.generate_bundle_document: bool = generate_bundle_document
        self.no_parallelization: bool = no_parallelization
        self.enable_mathjax: bool = enable_mathjax
        self.included_documents: bool = included_documents
        self.filter_requirements: Optional[str] = filter_requirements
        self.filter_sections: Optional[str] = filter_sections
        self.reqif_profile: Optional[str] = reqif_profile
        self.reqif_multiline_is_xhtml: bool = reqif_multiline_is_xhtml
        self.reqif_enable_mid: bool = reqif_enable_mid
        self.view: Optional[str] = view
        self.chromedriver: Optional[str] = chromedriver

    def get_path_to_config(self) -> str:
        # FIXME: The control flow can be improved.
        path_to_input_dir: str = self.input_paths[0]
        if os.path.isfile(path_to_input_dir):
            path_to_input_dir = os.path.dirname(path_to_input_dir)
        path_to_config = (
            self._config_path
            if self._config_path is not None
            else path_to_input_dir
        )
        return path_to_config

    def validate(self) -> None:
        for idx_, input_path_ in enumerate(self.input_paths.copy()):
            if not os.path.exists(input_path_):
                raise CLIValidationError(
                    f"Provided input path does not exist: {input_path_}"
                )
            if not os.path.isabs(input_path_):
                self.input_paths[idx_] = os.path.abspath(input_path_).rstrip(
                    "/"
                )
        if self._config_path is not None:
            if not os.path.exists(self._config_path):
                raise CLIValidationError(
                    "Provided path to a configuration file does not exist: "
                    f"{self._config_path}"
                )


class DumpGrammarCommandConfig:
    def __init__(self, output_file: str):
        self.output_file: str = output_file


class DiffCommandConfig:
    def __init__(
        self,
        path_to_lhs_tree: str,
        path_to_rhs_tree: str,
        output_dir: Optional[str],
    ):
        self.path_to_lhs_tree: str = path_to_lhs_tree
        self.path_to_rhs_tree: str = path_to_rhs_tree
        self.output_dir: Optional[str] = output_dir


class SDocArgsParser:
    def __init__(self, args: argparse.Namespace):
        self.args: argparse.Namespace = args

    def is_debug_mode(self) -> bool:
        return assert_cast(self.args.debug, bool)

    @property
    def is_about_command(self) -> bool:
        return str(self.args.command) == "about"

    @property
    def is_passthrough_command(self) -> bool:
        return str(self.args.command) == "passthrough"

    @property
    def is_export_command(self) -> bool:
        return str(self.args.command) == "export"

    @property
    def is_import_command_reqif(self) -> bool:
        return (
            str(self.args.command) == "import"
            and str(self.args.import_format) == "reqif"
        )

    @property
    def is_import_command_excel(self) -> bool:
        return (
            str(self.args.command) == "import"
            and str(self.args.import_format) == "excel"
        )

    @property
    def is_server_command(self) -> bool:
        return str(self.args.command) == "server"

    @property
    def is_dump_grammar_command(self) -> bool:
        return str(self.args.command) == "dump-grammar"

    @property
    def is_version_command(self) -> bool:
        return str(self.args.command) == "version"

    @property
    def is_diff_command(self) -> bool:
        return str(self.args.command) == "diff"

    @property
    def is_manage_autouid_command(self) -> bool:
        return (
            str(self.args.command) == "manage"
            and str(self.args.subcommand) == "auto-uid"
        )

    def get_export_config(self) -> ExportCommandConfig:
        project_title: Optional[str] = self.args.project_title

        return ExportCommandConfig(
            self.args.input_paths,
            self.args.output_dir,
            self.args.config,
            project_title,
            self.args.formats,
            self.args.fields,
            self.args.generate_bundle_document,
            self.args.no_parallelization,
            self.args.enable_mathjax,
            self.args.included_documents,
            self.args.filter_requirements,
            self.args.filter_sections,
            self.args.reqif_profile,
            self.args.reqif_multiline_is_xhtml,
            self.args.reqif_enable_mid,
            self.args.view,
            self.args.chromedriver,
        )

    def get_import_config_reqif(self, _: Any) -> ImportReqIFCommandConfig:
        return ImportReqIFCommandConfig(
            self.args.input_path,
            self.args.output_path,
            self.args.profile,
            self.args.reqif_enable_mid,
            self.args.reqif_import_markup,
        )

    def get_manage_autouid_config(self) -> ManageAutoUIDCommandConfig:
        return ManageAutoUIDCommandConfig(
            input_path=self.args.input_path,
            config_path=self.args.config,
            include_sections=self.args.include_sections,
        )

    def get_import_config_excel(self, _: Any) -> ImportExcelCommandConfig:
        return ImportExcelCommandConfig(
            self.args.input_path, self.args.output_path, self.args.parser
        )

    def get_server_config(self) -> ServerCommandConfig:
        return ServerCommandConfig(
            input_path=self.args.input_path,
            output_path=self.args.output_path,
            config_path=self.args.config,
            reload=self.args.reload,
            host=self.args.host,
            port=self.args.port,
        )

    def get_dump_grammar_config(self) -> DumpGrammarCommandConfig:
        return DumpGrammarCommandConfig(output_file=self.args.output_file)

    def get_diff_config(self) -> DiffCommandConfig:
        return DiffCommandConfig(
            path_to_lhs_tree=self.args.path_to_lhs_tree,
            path_to_rhs_tree=self.args.path_to_rhs_tree,
            output_dir=self.args.output_dir,
        )


def create_sdoc_args_parser(
    testing_args: Optional[argparse.Namespace] = None,
) -> SDocArgsParser:
    args = testing_args
    if not args:
        builder = CommandParserBuilder()
        parser = builder.build()
        args = parser.parse_args()
    return SDocArgsParser(args)



import re
from typing import Iterator, Optional, Set

from strictdoc.backend.sdoc.error_handling import StrictDocSemanticError
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.document_config import DocumentConfig
from strictdoc.backend.sdoc.models.document_grammar import (
    DocumentGrammar,
)
from strictdoc.backend.sdoc.models.document_view import DocumentView
from strictdoc.backend.sdoc.models.grammar_element import (
    GrammarElement,
    GrammarElementField,
    GrammarElementFieldMultipleChoice,
    GrammarElementFieldSingleChoice,
    GrammarElementFieldTag,
)
from strictdoc.backend.sdoc.models.model import RequirementFieldName
from strictdoc.backend.sdoc.models.node import SDocNode, SDocNodeField
from strictdoc.helpers.cast import assert_optional_cast

MULTIPLE_CHOICE_FIELD_KW = r"[a-zA-Z0-9\/\-|_ ]+"
MULTIPLE_CHOICE_FIELD_REGEX = re.compile(
    rf"^{MULTIPLE_CHOICE_FIELD_KW}(, {MULTIPLE_CHOICE_FIELD_KW})*$"
)


def multi_choice_regex_match(value: str) -> bool:
    """
    Validate MultipleChoice field value.

    Bug: Forward slashes in field values cause errors #2197,
    https://github.com/strictdoc-project/strictdoc/issues/2197
    """

    match = MULTIPLE_CHOICE_FIELD_REGEX.match(value)
    return match is not None


class SDocValidator:
    """
    Class that validates SDoc after textX parsing and processing steps.

    FIXME: In the future, all processing validation steps could be moved to this
           class as well. This would remove the need in the ParseContext class
           and simplify several other implementation details.
    """

    @staticmethod
    def validate_document(document: SDocDocument) -> None:
        assert isinstance(document, SDocDocument), document
        SDocValidator._validate_document_config(document)
        SDocValidator._validate_document_view(document)
        SDocValidator._validate_grammar(document)

    @staticmethod
    def validate_grammar_from_file(
        path_to_grammar: str, grammar_from_file: DocumentGrammar
    ) -> None:
        for grammar_element_ in grammar_from_file.elements:
            SDocValidator.validate_grammar_element(
                path_to_grammar, grammar_element_
            )

    @staticmethod
    def _validate_grammar(document: SDocDocument) -> None:
        assert document.meta is not None
        assert document.grammar is not None
        for grammar_element_ in document.grammar.elements:
            SDocValidator.validate_grammar_element(
                document.meta.input_doc_full_path, grammar_element_
            )

    @staticmethod
    def validate_grammar_element(
        path_to_grammar: str, grammar_element: GrammarElement
    ) -> None:
        # GrammarFromFile doesn't have a parent document.
        document: Optional[SDocDocument] = assert_optional_cast(
            grammar_element.parent.parent, SDocDocument
        )
        if document is not None and document.config.enable_mid:
            if "MID" not in grammar_element.fields_map:
                raise StrictDocSemanticError.grammar_element_has_no_mid_field(
                    grammar_element,
                    path_to_grammar,
                )

    @staticmethod
    def _validate_document_config(document: SDocDocument) -> None:
        document_config: DocumentConfig = document.config
        if document_config.default_view is not None:
            if document.view is None:
                raise StrictDocSemanticError.default_view_doesnt_exist(
                    document=document,
                    document_config=document_config,
                    default_view=document_config.default_view,
                )
            else:
                view_names = map(
                    lambda view_: view_.view_id, document.view.views
                )
                if document_config.default_view not in view_names:
                    raise StrictDocSemanticError.default_view_doesnt_exist(
                        document=document,
                        document_config=document_config,
                        default_view=document_config.default_view,
                    )

    @staticmethod
    def _validate_document_view(document: SDocDocument) -> None:
        assert document.grammar is not None

        document_view: Optional[DocumentView] = document.view
        if document_view is not None:
            for view in document_view.views:
                for tag in view.tags:
                    if (
                        tag.object_type
                        not in document.grammar.registered_elements
                    ):
                        raise StrictDocSemanticError.view_references_nonexisting_grammar_element(
                            document,
                            document_view,
                            view,
                            tag.object_type,
                        )

                    grammar_element: GrammarElement = (
                        document.grammar.elements_by_type[tag.object_type]
                    )
                    for field in tag.visible_fields:
                        if field.name not in grammar_element.fields_map:
                            raise StrictDocSemanticError.view_references_nonexisting_field(
                                document,
                                document_view,
                                view,
                                tag.object_type,
                                field.name,
                            )

    @staticmethod
    def validate_node(
        requirement: SDocNode,
        document_grammar: DocumentGrammar,
        path_to_sdoc_file: str,
        auto_uid_mode: bool = False,
    ) -> None:
        if requirement.node_type not in document_grammar.registered_elements:
            raise StrictDocSemanticError.unknown_node_type(
                requirement, path_to_sdoc_file
            )

        grammar_element: GrammarElement = document_grammar.elements_by_type[
            requirement.node_type
        ]

        if requirement.is_composite:
            if not grammar_element.property_is_composite:
                raise StrictDocSemanticError.composite_node_and_non_composite_element_mismatch(
                    requirement, path_to_sdoc_file
                )
        else:
            if grammar_element.property_is_composite:
                raise StrictDocSemanticError.non_composite_node_and_composite_element_composite(
                    requirement, path_to_sdoc_file
                )

        registered_fields: Set[str] = set(grammar_element.get_field_titles())

        for field_name in requirement.ordered_fields_lookup:
            if field_name not in registered_fields:
                raise StrictDocSemanticError.unregistered_field(
                    field_name=field_name,
                    requirement=requirement,
                    document_grammar=document_grammar,
                    path_to_sdoc_file=path_to_sdoc_file,
                )

        grammar_fields_iterator: Iterator[GrammarElementField] = iter(
            grammar_element.fields
        )
        requirement_field_iterator: Iterator[SDocNodeField] = (
            requirement.enumerate_fields()
        )

        requirement_field: Optional[SDocNodeField] = next(
            requirement_field_iterator, None
        )
        grammar_field: Optional[GrammarElementField] = next(
            grammar_fields_iterator, None
        )

        while True:
            try:
                valid_or_not_required_field = (
                    SDocValidator.validate_requirement_field(
                        requirement,
                        document_grammar,
                        requirement_field=requirement_field,
                        grammar_field=grammar_field,
                        path_to_sdoc_file=path_to_sdoc_file,
                        auto_uid_mode=auto_uid_mode,
                    )
                )
            except StopIteration:
                break
            if valid_or_not_required_field:
                assert requirement_field is not None
                # COMMENT can appear multiple times.
                if requirement_field.field_name == RequirementFieldName.COMMENT:
                    requirement_field = next(requirement_field_iterator, None)
                    if (
                        requirement_field is not None
                        and requirement_field.field_name
                        == RequirementFieldName.COMMENT
                    ):
                        continue
                    else:
                        grammar_field = next(grammar_fields_iterator, None)
                        continue
                grammar_field = next(grammar_fields_iterator, None)
                requirement_field = next(requirement_field_iterator, None)
            else:
                assert grammar_field is not None
                assert not grammar_field.required or (
                    grammar_field.title in ("MID", "UID") and auto_uid_mode
                )
                grammar_field = next(grammar_fields_iterator, None)

        # REFS validation.
        for reference in requirement.relations:
            if not grammar_element.has_relation_type_role(
                relation_type=reference.ref_type,
                relation_role=reference.role,
            ):
                raise StrictDocSemanticError.invalid_reference_type_item(
                    node=requirement,
                    reference_item=reference,
                    path_to_sdoc_file=path_to_sdoc_file,
                )

    @staticmethod
    def validate_requirement_field(
        requirement: SDocNode,
        document_grammar: DocumentGrammar,
        requirement_field: Optional[SDocNodeField],
        grammar_field: Optional[GrammarElementField],
        path_to_sdoc_file: str,
        auto_uid_mode: bool = True,
    ) -> bool:
        """
        Validate a single node field using its grammar element field definition.
        """
        if grammar_field is None:
            if requirement_field is None:
                # Both grammar and requirements fields are over.
                # Validation is done.
                raise StopIteration

            # Unexpected field outside grammar.
            raise StrictDocSemanticError.wrong_field_order(
                node=requirement,
                document_grammar=document_grammar,
                problematic_field=requirement_field,
                path_to_sdoc_file=path_to_sdoc_file,
            )

        # No more requirement fields. Checking if all remaining grammar fields
        # are non-required.
        if requirement_field is None:
            if grammar_field.required:
                raise StrictDocSemanticError.missing_required_field(
                    node=requirement,
                    grammar_field=grammar_field,
                    document_grammar=document_grammar,
                    path_to_sdoc_file=path_to_sdoc_file,
                )
            return False

        if grammar_field.title != requirement_field.field_name:
            if grammar_field.required:
                if (
                    grammar_field.title
                    not in requirement.ordered_fields_lookup.keys()
                ):
                    if grammar_field.title in ("MID", "UID") and auto_uid_mode:
                        return False

                    raise StrictDocSemanticError.missing_required_field(
                        node=requirement,
                        grammar_field=grammar_field,
                        document_grammar=document_grammar,
                        path_to_sdoc_file=path_to_sdoc_file,
                    )

                raise StrictDocSemanticError.wrong_field_order(
                    node=requirement,
                    document_grammar=document_grammar,
                    problematic_field=requirement_field,
                    path_to_sdoc_file=path_to_sdoc_file,
                )
            return False

        requirement_field_text_value = requirement_field.get_text_value()

        if isinstance(grammar_field, GrammarElementFieldSingleChoice):
            if requirement_field_text_value not in grammar_field.options:
                raise StrictDocSemanticError.invalid_choice_field(
                    node=requirement,
                    document_grammar=document_grammar,
                    requirement_field=requirement_field,
                    path_to_sdoc_file=path_to_sdoc_file,
                )

        elif isinstance(grammar_field, GrammarElementFieldMultipleChoice):
            if not multi_choice_regex_match(requirement_field_text_value):
                raise StrictDocSemanticError.not_comma_separated_choices(
                    node=requirement,
                    requirement_field=requirement_field,
                    path_to_sdoc_file=path_to_sdoc_file,
                )

            requirement_field_value_components = (
                requirement_field_text_value.split(", ")
            )
            for component in requirement_field_value_components:
                if component not in grammar_field.options:
                    raise StrictDocSemanticError.invalid_multiple_choice_field(
                        node=requirement,
                        document_grammar=document_grammar,
                        requirement_field=requirement_field,
                        path_to_sdoc_file=path_to_sdoc_file,
                    )

        elif isinstance(grammar_field, GrammarElementFieldTag):
            if not multi_choice_regex_match(requirement_field_text_value):
                raise StrictDocSemanticError.not_comma_separated_tag_field(
                    node=requirement,
                    requirement_field=requirement_field,
                    path_to_sdoc_file=path_to_sdoc_file,
                )

        return True



import re
from typing import List

REGEX_NAME_PART = r"[A-Za-z0-9-_\.]+"
REGEX_FILENAME = rf"{REGEX_NAME_PART}"
REGEX_FOLDER = rf"{REGEX_NAME_PART}"

REGEX_WILDCARD = REGEX_FILENAME
# FIXME: This does the job but most probably could be simplified.
REGEX_DOUBLE_WILDCARD = (
    rf"(\\|\/)?({REGEX_NAME_PART})?((\\|\/){REGEX_NAME_PART})*(\\|\/)?"
)

REGEX_MASK_VALIDATION = rf"[^(\\|\/)]{REGEX_DOUBLE_WILDCARD}"

MOST_COMMON_CHARACTERS = ("[", "]", "(", ")", "{", "}", "?", "+", "!")


def validate_mask(mask: str) -> None:
    if mask == "":
        raise SyntaxError("Path mask must not be empty.")

    if not mask[0].isalnum() and mask[0] != "*":
        raise SyntaxError(
            "Path mask must start with an alphanumeric character or "
            "a wildcard symbol '*'."
        )

    if "//" in mask or "\\\\" in mask:
        raise SyntaxError("Path mask must not contain double slashes.")

    if "***" in mask:
        raise SyntaxError("Invalid wildcard: '***'.")

    for regex_symbol in MOST_COMMON_CHARACTERS:
        if regex_symbol in mask:
            raise SyntaxError(
                f"Path mask must not contain any of the special characters: "
                f"{MOST_COMMON_CHARACTERS}."
            )


def compile_regex_mask(path_mask: str) -> str:
    regex_mask = path_mask
    regex_mask = regex_mask.replace(".", "\\.")
    regex_mask = regex_mask.replace("**", "XXXX")
    regex_mask = regex_mask.replace("*", REGEX_WILDCARD)
    regex_mask = regex_mask.replace("XXXX", REGEX_DOUBLE_WILDCARD)
    # Support files like "file1.py" and "./file1.py"
    regex_mask = "^" + "(\\.\\/)?" + regex_mask + "$"  # noqa: ISC003
    return regex_mask


class PathFilter:
    def __init__(
        self, filtered_paths: List[str], positive_or_negative: bool
    ) -> None:
        self.filtered_paths: List[str] = filtered_paths
        compiled_regex_masks: List[str] = []
        for filtered_path in filtered_paths:
            validate_mask(filtered_path)
            compiled_regex_masks.append(compile_regex_mask(filtered_path))

        self.master_regex: re.Pattern[str] = re.compile(
            "(" + "|".join(compiled_regex_masks) + ")"
        )
        self.positive_or_negative: bool = positive_or_negative

    def match(self, found_path: str) -> bool:
        # If this is an empty positive filter, we want to match everything
        # positively. If this is an empty negative filter, we want to match
        # everything negatively.
        if len(self.filtered_paths) == 0:
            return self.positive_or_negative

        portable_found_path = found_path.replace("\\", "/")
        regex_match = self.master_regex.match(portable_found_path)
        if regex_match is not None:
            return True

        return False



"""
@relation(SDOC-SRS-112, scope=file)
"""

from typing import Dict, Optional

from markupsafe import Markup

from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.document_grammar import DocumentGrammar
from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.export.html.generators.view_objects.traceability_matrix_view_object import (
    TraceabilityMatrixViewObject,
)
from strictdoc.export.html.html_templates import HTMLTemplates
from strictdoc.export.html.renderers.link_renderer import LinkRenderer
from strictdoc.export.html.renderers.markup_renderer import MarkupRenderer


class TraceabilityMatrixHTMLGenerator:
    @staticmethod
    def export(
        *,
        project_config: ProjectConfig,
        traceability_index: TraceabilityIndex,
        html_templates: HTMLTemplates,
    ) -> Markup:
        assert isinstance(html_templates, HTMLTemplates)

        known_relations: Dict[str, Dict[Optional[str], bool]] = {
            "Parent": {},
            "Child": {},
            "File": {},
        }

        discovered_relation_types = set()

        document_: SDocDocument
        for document_ in traceability_index.document_tree.document_list:
            assert document_.grammar is not None
            document_grammar: DocumentGrammar = document_.grammar
            for grammar_element_ in document_grammar.elements:
                for relation_ in grammar_element_.relations:
                    discovered_relation_types.add(relation_.relation_type)

                    bucket = known_relations[relation_.relation_type]

                    if relation_.relation_role not in bucket:
                        bucket[relation_.relation_role] = True

        for relation_type_ in list(known_relations.keys()):
            if relation_type_ not in discovered_relation_types:
                del known_relations[relation_type_]

        #
        # Validate that all config-provided relation tuples are actually present
        # in the existing documents.
        # A typical config entry may look like this:
        # [                                                   # noqa: ERA001
        #     ("Parent", None),                               # noqa: ERA001
        #     ("Parent", "Refines"),                          # noqa: ERA001
        #     ("Parent", "REQUIREMENT_FOR"),                  # noqa: ERA001
        #     ("File", None)                                  # noqa: ERA001
        # ]                                                   # noqa: ERA001
        #
        config_relation_tuples = (
            project_config.traceability_matrix_relation_columns
            if project_config.traceability_matrix_relation_columns is not None
            else []
        )
        for config_relation_tuple_ in config_relation_tuples:
            assert config_relation_tuple_[0] in known_relations
            bucket = known_relations[config_relation_tuple_[0]]
            assert config_relation_tuple_[1] in bucket

        #
        # After the config values have been validated, merge both the
        # config-provided list of relation tuples with the list that was
        # discovered from the existing documents.
        #
        known_relations_list = list(config_relation_tuples)
        for relation_type_, bucket_ in known_relations.items():
            for relation_role_, _ in bucket_.items():
                relation_tuple = (relation_type_, relation_role_)
                if relation_tuple not in known_relations_list:
                    known_relations_list.append(relation_tuple)

        link_renderer = LinkRenderer(
            root_path="", static_path=project_config.dir_for_sdoc_assets
        )
        markup_renderer = MarkupRenderer.create(
            "RST",
            traceability_index,
            link_renderer,
            html_templates,
            project_config,
            None,
        )

        view_object = TraceabilityMatrixViewObject(
            traceability_index=traceability_index,
            project_config=project_config,
            link_renderer=link_renderer,
            markup_renderer=markup_renderer,
            known_relations_list=known_relations_list,
        )
        return view_object.render_screen(html_templates.jinja_environment())



import os.path
import shutil
import subprocess
from pathlib import Path
from typing import Optional

from strictdoc.core.project_config import ProjectConfig
from strictdoc.helpers.timing import measure_performance


class GitClient:
    def __init__(self, path_to_git_root: str) -> None:
        assert os.path.isdir(path_to_git_root)
        self.path_to_git_root: str = path_to_git_root

    @staticmethod
    def create_repo_from_local_copy(
        revision: str, project_config: ProjectConfig
    ) -> "GitClient":
        with measure_performance(f"Copy Git repo: {revision}"):
            path_to_cwd = os.getcwd()
            if revision == "HEAD+":
                path_to_project_git_dir = os.path.join(path_to_cwd, ".git")
                assert os.path.isdir(path_to_project_git_dir)
                return GitClient(path_to_cwd)

            path_to_sandbox = os.path.join(
                project_config.get_path_to_cache_dir(), "git"
            )
            path_to_sandbox_git_repo = os.path.join(path_to_sandbox, revision)
            path_to_sandbox_git_repo_git = os.path.join(
                path_to_sandbox_git_repo, ".git"
            )
            if revision != "HEAD+" and os.path.exists(
                path_to_sandbox_git_repo_git
            ):
                git_client = GitClient(path_to_sandbox_git_repo)

                if git_client.is_clean_branch():
                    return git_client

            Path(path_to_sandbox).mkdir(parents=True, exist_ok=True)

            # Running git worktree add ... below results with "path already
            # exists" error if the destination folder already exists, even if
            # empty.
            if os.path.exists(path_to_sandbox_git_repo):
                shutil.rmtree(path_to_sandbox_git_repo)

            result = subprocess.run(
                [
                    "git",
                    "worktree",
                    "add",
                    "--force",
                    path_to_sandbox_git_repo,
                    revision,
                ],
                cwd=path_to_cwd,
                capture_output=True,
                text=True,
                check=False,
            )
            assert result.returncode == 0, result

            git_client = GitClient(path_to_sandbox_git_repo)

            if revision != "HEAD+":
                git_client.hard_reset(revision=revision)
                git_client.clean()

            return git_client

    def is_clean_branch(self) -> bool:
        #
        # https://unix.stackexchange.com/a/155077/77389
        #
        result = subprocess.run(
            ["git", "status", "--porcelain"],
            cwd=self.path_to_git_root,
            capture_output=True,
            text=True,
            check=False,
        )
        if result.returncode != 0:
            return False
        return result.stdout == ""

    def check_revision(self, revision: str) -> str:
        assert isinstance(revision, str)
        assert len(revision) > 0
        result = subprocess.run(
            ["git", "rev-parse", revision],
            cwd=self.path_to_git_root,
            capture_output=True,
            text=True,
            check=False,
        )
        if result.returncode == 0:
            return result.stdout.strip()
        raise LookupError(f"Non-existing revision: {revision}.")

    def hard_reset(self, revision: Optional[str] = None) -> None:
        reset_args = ["git", "reset", "--hard"]
        if revision is not None:
            reset_args.append(revision)
        result = subprocess.run(
            reset_args,
            cwd=self.path_to_git_root,
            capture_output=False,
            text=True,
            check=True,
        )
        assert result.returncode == 0, result

    def clean(self) -> None:
        result = subprocess.run(
            ["git", "clean", "-fd"],
            cwd=self.path_to_git_root,
            capture_output=False,
            text=True,
            check=True,
        )
        assert result.returncode == 0, result



import contextlib
from functools import wraps
from time import time
from typing import Callable, Iterator, TypeVar

from typing_extensions import ParamSpec

from strictdoc.helpers.math import round_up

P = ParamSpec("P")
R = TypeVar("R")


def timing_decorator(name: str) -> Callable[[Callable[P, R]], Callable[P, R]]:
    def timing_internal(func: Callable[P, R]) -> Callable[P, R]:
        @wraps(func)
        def wrap(*args: P.args, **kw: P.kwargs) -> R:
            print(f"Step '{name}' start.", flush=True)  # noqa: T201
            time_start = time()
            result = func(*args, **kw)
            time_end = time()
            print(  # noqa: T201
                f"Step '{name}' took: {round_up(time_end - time_start, 2)} sec.",
                flush=True,
            )
            return result

        return wrap

    return timing_internal


@contextlib.contextmanager
def measure_performance(title: str) -> Iterator[None]:
    time_start = time()
    yield
    time_end = time()

    time_diff = time_end - time_start
    padded_name = f"{title} ".ljust(60, ".")
    padded_time = f" {time_diff:0.2f}".rjust(6, ".")
    print(f"{padded_name}{padded_time}s", flush=True)  # noqa: T201



"""
@relation(SDOC-SRS-142, scope=file)
"""

from typing import List, Optional, Sequence

import tree_sitter_cpp
from tree_sitter import Language, Node, Parser

from strictdoc.backend.sdoc_source_code.constants import FunctionAttribute
from strictdoc.backend.sdoc_source_code.marker_parser import MarkerParser
from strictdoc.backend.sdoc_source_code.models.function import Function
from strictdoc.backend.sdoc_source_code.models.function_range_marker import (
    FunctionRangeMarker,
    RangeMarkerType,
)
from strictdoc.backend.sdoc_source_code.models.line_marker import LineMarker
from strictdoc.backend.sdoc_source_code.models.range_marker import (
    RangeMarker,
)
from strictdoc.backend.sdoc_source_code.models.source_file_info import (
    SourceFileTraceabilityInfo,
)
from strictdoc.backend.sdoc_source_code.parse_context import ParseContext
from strictdoc.backend.sdoc_source_code.processors.general_language_marker_processors import (
    function_range_marker_processor,
    line_marker_processor,
    range_marker_processor,
    source_file_traceability_info_processor,
)
from strictdoc.backend.sdoc_source_code.tree_sitter_helpers import (
    traverse_tree,
    ts_find_child_node_by_type,
    ts_find_child_nodes_by_type,
)
from strictdoc.helpers.cast import assert_cast
from strictdoc.helpers.file_stats import SourceFileStats


class SourceFileTraceabilityReader_C:
    def __init__(self, parse_nodes: bool = False) -> None:
        self.parse_nodes: bool = parse_nodes

    def read(
        self,
        input_buffer: bytes,
        file_path: Optional[str] = None,
    ) -> SourceFileTraceabilityInfo:
        assert isinstance(input_buffer, bytes)

        file_size = len(input_buffer)

        traceability_info = SourceFileTraceabilityInfo([])

        if file_size == 0:
            return traceability_info

        file_stats = SourceFileStats.create(input_buffer)
        parse_context = ParseContext(file_path, file_stats)

        # Works since Python 3.9 but we also lint this with mypy from Python 3.8.
        language_arg = tree_sitter_cpp.language()
        py_language = Language(  # type: ignore[call-arg, unused-ignore]
            language_arg
        )
        parser = Parser(py_language)  # type: ignore[call-arg, unused-ignore]

        tree = parser.parse(input_buffer)

        nodes = traverse_tree(tree)
        for node_ in nodes:
            function_name: str
            function_markers: List[FunctionRangeMarker]
            function_comment_node: Optional[Node]
            if node_.type == "translation_unit":
                if (
                    len(node_.children) > 0
                    and node_.children[0].type == "comment"
                    and (comment_node := node_.children[0])
                ):
                    if comment_node.text is not None:
                        comment_text = comment_node.text.decode("utf-8")
                        source_node = MarkerParser.parse(
                            comment_text,
                            node_.start_point[0] + 1,
                            # It is important that +1 is not present here because
                            # currently StrictDoc does not display the last empty line (\n is 10).
                            node_.end_point[0]
                            if input_buffer[-1] == 10
                            else node_.end_point[0] + 1,
                            node_.start_point[0] + 1,
                            parse_nodes=self.parse_nodes,
                        )
                        for marker_ in source_node.markers:
                            if not isinstance(marker_, FunctionRangeMarker):
                                continue
                            # At the top level, only accept the scope=file markers.
                            # Everything else will be handled by functions and classes.
                            if marker_.scope != RangeMarkerType.FILE:
                                continue
                            if isinstance(marker_, FunctionRangeMarker) and (
                                function_range_marker_ := marker_
                            ):
                                function_range_marker_processor(
                                    function_range_marker_, parse_context
                                )
                                traceability_info.markers.append(
                                    function_range_marker_
                                )

            elif node_.type in ("declaration", "field_declaration"):
                function_declarator_node = ts_find_child_node_by_type(
                    node_, "function_declarator"
                )

                # C++ reference declaration wrap the function declaration one time.
                if function_declarator_node is None:
                    # Example: "TrkVertex& operator-=(const TrkVertex& c);".
                    reference_declarator_node = ts_find_child_node_by_type(
                        node_, "reference_declarator"
                    )
                    if reference_declarator_node is None:
                        continue

                    function_declarator_node = ts_find_child_node_by_type(
                        reference_declarator_node, "function_declarator"
                    )
                    if function_declarator_node is None:
                        continue

                # For normal C functions the identifier is "identifier".
                # For C++, there are:
                # Class function declarations: bool CanSend(const CanFrame &frame);         # noqa: ERA001
                # Operators:                   TrkVertex& operator-=(const TrkVertex& c);   # noqa: ERA001
                # Destructors:                 ~TrkVertex();                                # noqa: ERA001
                function_identifier_node = self._get_function_name_node(
                    function_declarator_node,
                )
                if function_identifier_node is None:
                    continue

                if function_identifier_node.text is None:
                    continue

                assert function_identifier_node.text is not None, node_.text
                function_display_name = function_identifier_node.text.decode(
                    "utf8"
                )

                assert function_declarator_node.text is not None, node_.text
                function_name = function_declarator_node.text.decode("utf8")
                assert function_name is not None, node_.text

                parent_names = self.get_node_ns(node_)
                if len(parent_names) > 0:
                    function_name = (
                        f"{'::'.join(parent_names)}::{function_name}"
                    )
                    function_display_name = (
                        f"{'::'.join(parent_names)}::{function_display_name}"
                    )

                function_attributes = {FunctionAttribute.DECLARATION}
                for specifier_node_ in ts_find_child_nodes_by_type(
                    node_, "storage_class_specifier"
                ):
                    if specifier_node_.text == b"static":
                        function_attributes.add(FunctionAttribute.STATIC)

                function_markers = []
                function_comment_node = None
                if (
                    node_.prev_sibling is not None
                    and node_.prev_sibling.type == "comment"
                ):
                    function_comment_node = node_.prev_sibling
                    assert function_comment_node.text is not None, node_.text
                    function_comment_text = function_comment_node.text.decode(
                        "utf8"
                    )

                    function_last_line = node_.end_point[0] + 1

                    source_node = MarkerParser.parse(
                        function_comment_text,
                        function_comment_node.start_point[0] + 1,
                        function_last_line,
                        function_comment_node.start_point[0] + 1,
                        entity_name=function_display_name,
                        parse_nodes=self.parse_nodes,
                    )
                    for marker_ in source_node.markers:
                        if isinstance(marker_, FunctionRangeMarker) and (
                            function_range_marker_ := marker_
                        ):
                            function_range_marker_processor(
                                function_range_marker_, parse_context
                            )
                            traceability_info.markers.append(
                                function_range_marker_
                            )
                            function_markers.append(marker_)

                # The function range includes the top comment if it exists.
                new_function = Function(
                    parent=traceability_info,
                    name=function_name,
                    display_name=function_display_name,
                    line_begin=function_comment_node.start_point[0] + 1
                    if function_comment_node is not None
                    else node_.range.start_point[0] + 1,
                    line_end=node_.range.end_point[0] + 1,
                    child_functions=[],
                    markers=function_markers,
                    attributes=function_attributes,
                )
                traceability_info.functions.append(new_function)

            elif node_.type == "function_definition":
                function_name = ""

                function_declarator_node = ts_find_child_node_by_type(
                    node_, "function_declarator"
                )
                # C++ reference declaration wrap the function declaration one time.
                if function_declarator_node is None:
                    # Example: Foo& Foo::operator+(const Foo& c) { return *this; }
                    reference_declarator_node = ts_find_child_node_by_type(
                        node_, "reference_declarator"
                    )
                    if reference_declarator_node is None:
                        continue

                    function_declarator_node = ts_find_child_node_by_type(
                        reference_declarator_node, "function_declarator"
                    )
                    if function_declarator_node is None:
                        continue

                assert function_declarator_node is not None, node_.text

                assert function_declarator_node.text is not None, node_.text
                function_name = function_declarator_node.text.decode("utf8")

                identifier_node = self._get_function_name_node(
                    function_declarator_node
                )
                if identifier_node is None:
                    raise NotImplementedError(function_declarator_node)

                assert identifier_node.text is not None, node_.text
                function_display_name = identifier_node.text.decode("utf8")

                assert function_name is not None, node_.text
                parent_names = self.get_node_ns(node_)
                if len(parent_names) > 0:
                    function_name = (
                        f"{'::'.join(parent_names)}::{function_name}"
                    )
                    function_display_name = (
                        f"{'::'.join(parent_names)}::{function_display_name}"
                    )

                # FIXME: Special hack for Google Test macros TEST, TEST_F, TEST_P.
                if function_name.startswith("TEST"):
                    function_display_name = function_name

                function_markers = []
                function_comment_node = None
                function_comment_text = None

                # In the condition below, it is important that the comment is
                # considered a function comment only if it there are no empty
                # lines between the comment and function.
                if (
                    node_.prev_sibling is not None
                    and node_.prev_sibling.type == "comment"
                    and (node_.prev_sibling.end_point[0] + 1)
                    == node_.start_point[0]
                ):
                    function_comment_node = node_.prev_sibling
                    assert function_comment_node.text is not None, node_.text
                    function_comment_text = function_comment_node.text.decode(
                        "utf8"
                    )

                    function_last_line = node_.end_point[0] + 1

                    source_node = MarkerParser.parse(
                        function_comment_text,
                        function_comment_node.start_point[0] + 1,
                        function_last_line,
                        function_comment_node.start_point[0] + 1,
                        entity_name=function_display_name,
                        parse_nodes=self.parse_nodes,
                    )
                    traceability_info.source_nodes.append(source_node)
                    for marker_ in source_node.markers:
                        if isinstance(marker_, FunctionRangeMarker):
                            function_range_marker_processor(
                                marker_, parse_context
                            )
                            traceability_info.markers.append(marker_)
                            function_markers.append(marker_)

                # The function range includes the top comment if it exists.
                new_function = Function(
                    parent=traceability_info,
                    name=function_name,
                    display_name=function_display_name,
                    line_begin=function_comment_node.start_point[0] + 1
                    if function_comment_node is not None
                    else node_.range.start_point[0] + 1,
                    line_end=node_.range.end_point[0] + 1,
                    child_functions=[],
                    markers=function_markers,
                    attributes={FunctionAttribute.DEFINITION},
                )
                traceability_info.functions.append(new_function)
                if len(function_markers) > 0:
                    traceability_info.ng_map_names_to_markers[function_name] = (
                        # FIXME: Cannot win the fight with mypy without assert_cast.
                        assert_cast(function_markers, list)
                    )
                    traceability_info.ng_map_names_to_definition_functions[
                        function_name
                    ] = new_function
            elif node_.type == "comment":
                #
                # FIXME: Here parsing of function comments can happen as well
                #        but this time the focus is ONLY on range and line markers.
                #        The case which is handled here is when a user adds a
                #        range_start marker in a function comment.
                #        It is not good that parsing of function comments
                #        happens twice.
                #

                assert node_.text is not None, (
                    f"Comment without a text: {node_}"
                )

                node_text_string = node_.text.decode("utf8")

                source_node = MarkerParser.parse(
                    node_text_string,
                    node_.start_point[0] + 1,
                    node_.end_point[0] + 1,
                    node_.start_point[0] + 1,
                    parse_nodes=False,
                )

                for marker_ in source_node.markers:
                    if isinstance(marker_, RangeMarker) and (
                        range_marker_ := marker_
                    ):
                        range_marker_processor(range_marker_, parse_context)
                    elif isinstance(marker_, LineMarker) and (
                        line_marker_ := marker_
                    ):
                        line_marker_processor(line_marker_, parse_context)
                    else:
                        pass
            else:
                pass

        source_file_traceability_info_processor(
            traceability_info, parse_context
        )

        traceability_info.ng_map_reqs_to_markers = (
            parse_context.map_reqs_to_markers
        )

        return traceability_info

    def read_from_file(self, file_path: str) -> SourceFileTraceabilityInfo:
        with open(file_path, "rb") as file:
            sdoc_content = file.read()
            sdoc = self.read(sdoc_content, file_path=file_path)
            return sdoc

    @staticmethod
    def _get_function_name_node(
        function_declarator_node: Node,
    ) -> Optional[Node]:
        assert function_declarator_node.type == "function_declarator"
        function_identifier_node = ts_find_child_node_by_type(
            function_declarator_node,
            node_type=(
                "identifier",
                "field_identifier",
                "operator_name",
                "destructor_name",
                "qualified_identifier",
            ),
        )
        return function_identifier_node

    @staticmethod
    def get_node_ns(node: Node) -> Sequence[str]:
        """
        Walk up the tree and find parent classes.
        """
        parent_scopes = []
        cursor: Optional[Node] = node
        while cursor is not None:
            if cursor.type == "class_specifier" and len(cursor.children) > 1:
                second_node_or_none = cursor.children[1]
                if (
                    second_node_or_none.type == "type_identifier"
                    and second_node_or_none.text is not None
                ):
                    parent_class_name = second_node_or_none.text.decode("utf8")
                    parent_scopes.append(parent_class_name)

            cursor = cursor.parent

        parent_scopes.reverse()
        return parent_scopes



import strictdoc


class AboutCommand:
    @staticmethod
    def execute() -> None:
        print("=============")  # noqa: T201
        print("= StrictDoc =")  # noqa: T201
        print("=============")  # noqa: T201
        print(  # noqa: T201
            "Purpose: Software for writing technical requirements specifications."
        )
        print(f"Version: {strictdoc.__version__}")  # noqa: T201
        print(  # noqa: T201
            "Docs:    https://strictdoc.readthedocs.io/en/stable/"
        )
        print(  # noqa: T201
            "Github:  https://github.com/strictdoc-project/strictdoc"
        )
        print("License: Apache 2")  # noqa: T201



import sys

from strictdoc.backend.sdoc.errors.document_tree_error import DocumentTreeError
from strictdoc.backend.sdoc.writer import SDWriter
from strictdoc.core.analyzers.document_stats import (
    DocumentStats,
    DocumentTreeStats,
)
from strictdoc.core.analyzers.document_uid_analyzer import DocumentUIDAnalyzer
from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.core.traceability_index_builder import TraceabilityIndexBuilder
from strictdoc.helpers.parallelizer import Parallelizer
from strictdoc.helpers.string import (
    create_safe_acronym,
)


class ManageAutoUIDCommand:
    @staticmethod
    def execute(
        *, project_config: ProjectConfig, parallelizer: Parallelizer
    ) -> None:
        # FIXME: Traceability Index is coupled with HTML output.
        project_config.export_output_html_root = "NOT_RELEVANT"

        try:
            traceability_index: TraceabilityIndex = (
                TraceabilityIndexBuilder.create(
                    project_config=project_config,
                    parallelizer=parallelizer,
                )
            )
        except DocumentTreeError as exc:
            print(exc.to_print_message())  # noqa: T201
            sys.exit(1)

        document_tree_stats: DocumentTreeStats = (
            DocumentUIDAnalyzer.analyze_document_tree(traceability_index)
        )

        if project_config.autouuid_include_sections:
            document_stats_: DocumentStats
            for document_stats_ in document_tree_stats.single_document_stats:
                document_acronym = create_safe_acronym(
                    document_stats_.document.title
                )
                for section in document_stats_.sections_without_uid:
                    auto_uid = document_tree_stats.get_auto_section_uid(
                        document_acronym, section
                    )
                    section.reserved_uid = auto_uid

        for (
            prefix,
            prefix_requirements,
        ) in document_tree_stats.requirements_per_prefix.items():
            next_number = document_tree_stats.get_next_requirement_uid_number(
                prefix
            )

            for requirement in prefix_requirements.requirements_no_uid:
                requirement_prefix = requirement.get_prefix()
                requirement_uid = f"{requirement_prefix}{next_number}"
                requirement.set_field_value(
                    field_name="UID",
                    form_field_index=0,
                    value=requirement_uid,
                )
                next_number += 1

        for document in traceability_index.document_tree.document_list:
            assert document.meta is not None

            # Most recently, we parse JUnit XML or Gcov JSON files.
            # These must not be written back.
            if not document.meta.document_filename.endswith(".sdoc"):
                continue

            document_content = SDWriter(project_config).write(document)
            document_meta = document.meta
            with open(
                document_meta.input_doc_full_path, "w", encoding="utf8"
            ) as output_file:
                output_file.write(document_content)



"""
@relation(SDOC-SRS-134, scope=file)
"""

import os
from pathlib import Path
from typing import Dict, List

import xlsxwriter
from xlsxwriter.workbook import Workbook
from xlsxwriter.worksheet import Worksheet

from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.grammar_element import ReferenceType
from strictdoc.backend.sdoc.models.model import (
    SDocIteratedElementIF,
)
from strictdoc.backend.sdoc.models.node import SDocNode
from strictdoc.backend.sdoc.models.reference import (
    FileReference,
    ParentReqReference,
)
from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.helpers.cast import assert_cast

EXCEL_SHEET_NAME = "Requirements"
MAX_WIDTH = 75
HEADER_MARGIN = 3
MAX_WIDTH_KEY = "max_width"
COLUMN_HEADER_KEY = "header"
PARENT_COLUMN_HEADER_LABEL = "PARENT"
EXPORT_COLUMNS = [
    {
        "name": "uid",
        "header": "UID",
    },
    {
        "name": "statement",
        "header": "STATEMENT",
    },
]


class ExcelGenerator:
    @staticmethod
    def export_tree(
        traceability_index: TraceabilityIndex,
        output_excel_root: str,
        project_config: ProjectConfig,
    ) -> None:
        Path(output_excel_root).mkdir(parents=True, exist_ok=True)

        document: SDocDocument
        for document in traceability_index.document_tree.document_list:
            assert document.meta is not None

            document_out_file_name = (
                f"{document.meta.document_filename_base}.xlsx"
            )
            document_out_file = os.path.join(
                output_excel_root, document_out_file_name
            )

            ExcelGenerator._export_single_document(
                document, traceability_index, document_out_file, project_config
            )

    @staticmethod
    def _export_single_document(
        document: SDocDocument,
        traceability_index: TraceabilityIndex,
        document_out_file: str,
        project_config: ProjectConfig,
    ) -> None:
        with xlsxwriter.Workbook(document_out_file) as workbook:
            worksheet = workbook.add_worksheet(name=EXCEL_SHEET_NAME)
            workbook.set_properties(
                {
                    "title": project_config.project_title,
                    "comments": "Created with StrictDoc.",
                }
            )

            # Header row.
            row = 1

            fields = project_config.excel_export_fields

            # FIXME: Check if all fields are defined by the DocumentGrammar.
            assert fields is not None
            columns = ExcelGenerator._init_columns_width(fields)

            document_iterator = traceability_index.get_document_iterator(
                document
            )
            all_nodes = list(
                map(
                    lambda node_with_context_: node_with_context_[0],
                    document_iterator.all_content(print_fragments=False),
                )
            )
            req_uid_rows = ExcelGenerator._lookup_refs(all_nodes)

            if len(req_uid_rows):
                for node, _ in traceability_index.get_document_iterator(
                    document
                ).all_content(print_fragments=False):
                    if (
                        not isinstance(node, SDocNode)
                        or node.reserved_uid is None
                    ):
                        # Only export the requirements with UID.
                        continue

                    for idx, field in enumerate(fields, start=0):
                        field_uc = field.upper()

                        # Special treatment for ParentReqReference and Comments.
                        if field_uc in (
                            "RELATIONS:PARENT",
                            "PARENT",
                            "PARENTS",
                        ):
                            parent_refs = node.get_requirement_references(
                                ReferenceType.PARENT
                            )
                            if len(parent_refs) > 0:
                                # FIXME: Allow multiple parent refs.
                                ref = assert_cast(
                                    parent_refs[0], ParentReqReference
                                )
                                columns[field][MAX_WIDTH_KEY] = max(
                                    len(ref.ref_uid),
                                    columns[field][MAX_WIDTH_KEY],
                                )
                                if ref.ref_uid in req_uid_rows:
                                    worksheet.write_url(
                                        row,
                                        idx,
                                        (
                                            "internal:"
                                            f"'{EXCEL_SHEET_NAME}'"
                                            f"!A{req_uid_rows[ref.ref_uid]}"
                                        ),
                                        string=ref.ref_uid,
                                    )
                                else:
                                    worksheet.write(row, idx, ref.ref_uid)
                        elif field_uc in ("COMMENT", "COMMENTS"):
                            # Using a transition marker to separate multiple
                            # comments.
                            comment_fields = node.get_comment_fields()
                            if len(comment_fields) > 0:
                                comment_row_value: str = ""
                                for comment_field_ in comment_fields:
                                    if len(comment_row_value) > 0:
                                        comment_row_value += "\n----------\n"
                                    comment_row_value += (
                                        comment_field_.get_text_value()
                                    )
                                worksheet.write(row, idx, comment_row_value)
                                if (
                                    comment_row_value
                                    and len(comment_row_value)
                                    > columns[field][MAX_WIDTH_KEY]
                                ):
                                    columns[field][MAX_WIDTH_KEY] = len(
                                        comment_row_value
                                    )
                        elif field_uc == "RELATIONS":
                            if len(node.relations) > 0:
                                relations_components = []
                                # Using a transition marker to separate
                                # multiple references.
                                for relation_ in node.relations:
                                    if isinstance(
                                        relation_, ParentReqReference
                                    ):
                                        relations_components.append(
                                            relation_.ref_type
                                            + ": "
                                            + relation_.ref_uid
                                        )
                                    elif isinstance(relation_, FileReference):
                                        relations_components.append(
                                            relation_.ref_type
                                            + ": "
                                            + relation_.get_posix_path()
                                        )
                                relations_row_value: str = (
                                    "\n----------\n".join(relations_components)
                                )
                                worksheet.write(row, idx, relations_row_value)
                                value_len = len(relations_row_value)
                                columns[field][MAX_WIDTH_KEY] = max(
                                    value_len, columns[field][MAX_WIDTH_KEY]
                                )
                        elif field_uc in node.ordered_fields_lookup.keys():
                            req_field = node.ordered_fields_lookup[field_uc][0]
                            value: str = req_field.get_text_value()
                            worksheet.write(row, idx, value)
                            value_len = len(value)
                            columns[field][MAX_WIDTH_KEY] = max(
                                value_len, columns[field][MAX_WIDTH_KEY]
                            )
                        elif hasattr(node, "special_fields"):
                            if len(node.special_fields):
                                for special_field in node.special_fields:
                                    if (
                                        special_field.field_name.upper()
                                        == field_uc
                                    ):
                                        value = special_field.field_value
                                        worksheet.write(row, idx, value)
                                        if (
                                            value
                                            and len(value)
                                            > columns[field][MAX_WIDTH_KEY]
                                        ):
                                            columns[field][MAX_WIDTH_KEY] = len(
                                                value
                                            )
                                        break

                    row += 1

                # Add a table around all this data, allowing filtering and
                # ordering in Excel.
                worksheet.add_table(
                    0,
                    0,
                    row - 1,
                    len(fields) - 1,
                    {"columns": ExcelGenerator._init_headers(fields)},
                )

                # Enforce columns width.
                ExcelGenerator._set_columns_width(
                    workbook, worksheet, columns, fields
                )
            else:
                # No requirement with UID.
                print(  # noqa: T201
                    "No requirement with UID, nothing to export into excel"
                )

        if row == 1:
            os.unlink(document_out_file)

    @staticmethod
    def _lookup_refs(
        document_contents: List[SDocIteratedElementIF],
    ) -> Dict[str, int]:
        refs: Dict[str, int] = {}
        row = 1

        for content_node in document_contents:
            if isinstance(content_node, SDocNode):
                if content_node.reserved_uid:
                    # Only export the requirements with uid, allowing tracking.
                    row += 1
                    refs[content_node.reserved_uid] = row

        return refs

    @staticmethod
    def _init_columns_width(fields: List[str]) -> Dict[str, Dict[str, int]]:
        columns: Dict[str, Dict[str, int]] = {}

        for field in fields:
            columns[field] = {}
            columns[field][MAX_WIDTH_KEY] = len(field) + HEADER_MARGIN
        return columns

    @staticmethod
    def _set_columns_width(
        workbook: Workbook,
        worksheet: Worksheet,
        columns: Dict[str, Dict[str, int]],
        fields: List[str],
    ) -> None:
        cell_format_text_wrap = workbook.add_format()
        cell_format_text_wrap.set_text_wrap()

        for idx, field in enumerate(fields, start=0):
            if columns[field][MAX_WIDTH_KEY] > MAX_WIDTH:
                worksheet.set_column(idx, idx, MAX_WIDTH, cell_format_text_wrap)
            else:
                worksheet.set_column(idx, idx, columns[field][MAX_WIDTH_KEY])

    @staticmethod
    def _init_headers(fields: List[str]) -> List[Dict[str, str]]:
        headers: List[Dict[str, str]] = []

        for field in fields:
            headers.append({"header": field.upper()})

        return headers



"""
@relation(SDOC-SRS-34, SDOC-SRS-141, scope=file)
"""

from typing import Dict, List, Optional, Tuple, Union

from lark import ParseTree, Token, Tree

from strictdoc.backend.sdoc_source_code.comment_parser.marker_lexer import (
    MarkerLexer,
)
from strictdoc.backend.sdoc_source_code.helpers.comment_preprocessor import (
    preprocess_source_code_comment,
)
from strictdoc.backend.sdoc_source_code.models.function_range_marker import (
    FunctionRangeMarker,
)
from strictdoc.backend.sdoc_source_code.models.line_marker import LineMarker
from strictdoc.backend.sdoc_source_code.models.range_marker import (
    RangeMarker,
)
from strictdoc.backend.sdoc_source_code.models.requirement_marker import Req
from strictdoc.backend.sdoc_source_code.models.source_node import SourceNode


class MarkerParser:
    @staticmethod
    def parse(
        input_string: str,
        line_start: int,
        line_end: int,
        comment_line_start: int,
        entity_name: Optional[str] = None,
        col_offset: int = 0,
        parse_nodes: bool = False,
    ) -> SourceNode:
        """
        Parse relation markers from source file comments.

        Before the actual parsing, the function removes all code comment symbols
        such as /** ... */ or /// Doxygen comments or Python

        The line start/end indicate a range where the comment is located in the
        source file.
        The comment_line_start parameter maybe the same as line start but can
        also be different if the comment is part of a Python or C function in
        which case the comment_line_start will be several lines below the actual
        start of the range.
        """

        node_fields: Dict[str, str] = {}
        source_node: SourceNode = SourceNode(entity_name)

        input_string = preprocess_source_code_comment(input_string)

        tree: ParseTree = MarkerLexer.parse(
            input_string, parse_nodes=parse_nodes
        )

        for element_ in tree.children:
            if not isinstance(element_, Tree):
                continue

            if element_.data == "relation_marker":
                relation_markers = MarkerParser._parse_relation_marker(
                    element_,
                    line_start,
                    line_end,
                    comment_line_start,
                    entity_name,
                    col_offset,
                )
                source_node.markers.extend(relation_markers)

            elif element_.data == "node_field":
                node_name, node_value = MarkerParser._parse_node_field(
                    element_,
                )
                node_fields[node_name] = node_value
            else:
                raise AssertionError

        if len(node_fields) > 0:
            source_node.fields = node_fields

        return source_node

    @staticmethod
    def _parse_relation_marker(
        element_: Tree[Token],
        line_start: int,
        line_end: int,
        comment_line_start: int,
        entity_name: Optional[str] = None,
        col_offset: int = 0,
    ) -> List[Union[FunctionRangeMarker, RangeMarker, LineMarker]]:
        markers: List[Union[FunctionRangeMarker, RangeMarker, LineMarker]] = []

        relation_uid_elements = []
        relation_scope_element: Optional[Tree[Token]] = None
        relation_role_element: Optional[Tree[Token]] = None
        for relation_marker_element_ in element_.children:
            assert isinstance(relation_marker_element_, Tree)
            if relation_marker_element_.data == "relation_node_uid":
                relation_uid_elements.append(relation_marker_element_)
            elif relation_marker_element_.data == "relation_scope":
                relation_scope_element = relation_marker_element_
            elif relation_marker_element_.data == "relation_role":
                relation_role_element = relation_marker_element_
            else:
                raise NotImplementedError

        assert len(relation_uid_elements) > 0
        assert relation_scope_element is not None

        assert isinstance(relation_scope_element.children[0], Token)
        relation_scope = relation_scope_element.children[0].value

        relation_role = None
        if relation_role_element is not None:
            assert isinstance(relation_role_element.children[0], Token)
            relation_role = relation_role_element.children[0].value

        requirements = []
        used_uids = set()

        for relation_uid_token_ in relation_uid_elements:
            assert isinstance(relation_uid_token_.children[0], Token)
            relation_uid = relation_uid_token_.children[0].value
            if relation_uid in used_uids:
                raise ValueError(
                    f"@relation marker contains duplicate node UIDs: ['{relation_uid}']."
                )
            used_uids.add(relation_uid)

            assert relation_uid_token_.children[0].line is not None
            requirement = Req(None, relation_uid)
            requirement.ng_source_line = (
                comment_line_start + relation_uid_token_.children[0].line - 1
            )
            requirement.ng_source_column = relation_uid_token_.children[
                0
            ].column
            requirements.append(requirement)

        if relation_scope in ("file", "class", "function"):
            function_marker = FunctionRangeMarker(
                None, requirements, scope=relation_scope, role=relation_role
            )
            function_marker.ng_source_line_begin = (
                comment_line_start + element_.meta.line - 1
            )
            function_marker.ng_source_column_begin = (
                element_.meta.column + col_offset
            )
            function_marker.ng_range_line_begin = line_start
            function_marker.ng_range_line_end = line_end
            if relation_scope == "file":
                function_marker.set_description("entire file")
            elif relation_scope == "function":
                function_marker.set_description(f"function {entity_name}()")
            elif relation_scope == "class":
                function_marker.set_description(f"class {entity_name}")
            markers.append(function_marker)
        elif relation_scope in ("range_start", "range_end"):
            start_or_end = relation_scope == "range_start"
            range_marker = RangeMarker(
                None,
                "[" if start_or_end else "[/",
                requirements,
                role=relation_role,
            )
            range_marker.ng_source_line_begin = (
                comment_line_start + element_.meta.line - 1
            )
            range_marker.ng_source_column_begin = (
                element_.meta.column + col_offset
            )
            range_marker.ng_range_line_begin = line_start
            range_marker.ng_range_line_end = line_end
            range_marker.ng_new_relation_keyword = True
            markers.append(range_marker)
        elif relation_scope == "line":
            line_marker = LineMarker(None, requirements, role=relation_role)
            line_marker.ng_source_line_begin = (
                comment_line_start + element_.meta.line - 1
            )
            line_marker.ng_source_column_begin = (
                element_.meta.column + col_offset
            )
            line_marker.ng_range_line_begin = line_start
            line_marker.ng_range_line_end = line_end
            markers.append(line_marker)
        else:
            raise NotImplementedError

        return markers

    @staticmethod
    def _parse_node_field(
        element_: Tree[Token],
    ) -> Tuple[str, str]:
        node_name_node = element_.children[0]
        assert isinstance(node_name_node, Tree)
        assert node_name_node.data == "node_name"
        assert isinstance(node_name_node.children[0], Token)
        node_name = node_name_node.children[0].value

        node_value_node = element_.children[1]
        assert isinstance(node_value_node, Tree)
        assert node_value_node.data == "node_multiline_value"

        processed_node_values = []
        for node_value_component_ in node_value_node.children:
            assert isinstance(node_value_component_, Token)
            processed_node_value = node_value_component_.value.strip()
            if "\\n\\n" in processed_node_value:
                processed_node_value = processed_node_value.replace(
                    "\\n\\n", ""
                )

            processed_node_values.append(processed_node_value)

        node_value = "\n".join(processed_node_values)

        return node_name, node_value



"""
@relation(SDOC-SRS-142, scope=file)
"""

from itertools import islice
from typing import List, Optional, Sequence

import tree_sitter_python
from tree_sitter import Language, Node, Parser

from strictdoc.backend.sdoc_source_code.marker_parser import MarkerParser
from strictdoc.backend.sdoc_source_code.models.function import Function
from strictdoc.backend.sdoc_source_code.models.function_range_marker import (
    FunctionRangeMarker,
)
from strictdoc.backend.sdoc_source_code.models.line_marker import LineMarker
from strictdoc.backend.sdoc_source_code.models.range_marker import (
    RangeMarker,
)
from strictdoc.backend.sdoc_source_code.models.source_file_info import (
    RelationMarkerType,
    SourceFileTraceabilityInfo,
)
from strictdoc.backend.sdoc_source_code.parse_context import ParseContext
from strictdoc.backend.sdoc_source_code.processors.general_language_marker_processors import (
    function_range_marker_processor,
    line_marker_processor,
    range_marker_processor,
    source_file_traceability_info_processor,
)
from strictdoc.backend.sdoc_source_code.tree_sitter_helpers import traverse_tree
from strictdoc.helpers.file_stats import SourceFileStats


class SourceFileTraceabilityReader_Python:
    def read(
        self, input_buffer: bytes, file_path: Optional[str] = None
    ) -> SourceFileTraceabilityInfo:
        assert isinstance(input_buffer, bytes)

        file_size = len(input_buffer)

        traceability_info = SourceFileTraceabilityInfo([])

        if file_size == 0:
            return traceability_info

        file_stats = SourceFileStats.create(input_buffer)
        parse_context = ParseContext(file_path, file_stats)

        # Works since Python 3.9 but we also lint this with mypy from Python 3.8.
        language_arg = tree_sitter_python.language()
        py_language = Language(  # type: ignore[call-arg, unused-ignore]
            language_arg
        )
        parser = Parser(py_language)  # type: ignore[call-arg, unused-ignore]

        tree = parser.parse(input_buffer)

        functions_stack: List[Function] = []

        nodes = traverse_tree(tree)
        map_function_to_node = {}
        for node_ in nodes:
            if node_.type == "module":
                function = Function(
                    parent=traceability_info,
                    name="module",
                    display_name="module",
                    line_begin=node_.start_point[0] + 1,
                    line_end=node_.end_point[0] + 1,
                    child_functions=[],
                    markers=[],
                    attributes=set(),
                )
                functions_stack.append(function)
                map_function_to_node[function] = node_
                if len(node_.children) > 0:
                    # Look for the docstring within the first 30 children (arbitrary chosen limit)
                    # so that we dont miss it if the file starts with comments (#!, encoding marker, etc...).
                    first_match = next(
                        (
                            child
                            for child in islice(node_.children, 30)
                            if child.type == "expression_statement"
                            and len(child.children) > 0
                            and child.children[0].type == "string"
                        ),
                        None,
                    )

                    if first_match is not None:
                        block_comment = first_match.children[0]

                        # String contains of three parts:
                        # (string_start string_content string_end)
                        string_content = block_comment.children[1]
                        assert string_content.text is not None

                        block_comment_text = string_content.text.decode("utf-8")
                        source_node = MarkerParser.parse(
                            block_comment_text,
                            node_.start_point[0] + 1,
                            # It is important that +1 is not present here because
                            # currently StrictDoc does not display the last empty line (\n is 10).
                            node_.end_point[0]
                            if input_buffer[-1] == 10
                            else node_.end_point[0] + 1,
                            string_content.start_point[0] + 1,
                        )
                        for marker_ in source_node.markers:
                            if isinstance(marker_, FunctionRangeMarker) and (
                                function_range_marker_ := marker_
                            ):
                                function_range_marker_processor(
                                    function_range_marker_, parse_context
                                )
                                traceability_info.markers.append(
                                    function_range_marker_
                                )

            elif node_.type in ("class_definition", "function_definition"):
                function_name: str = ""
                function_block: Optional[Node] = None

                for child_ in node_.children:
                    if child_.type == "identifier":
                        if child_.text is not None:
                            function_name = child_.text.decode("utf-8")
                    if child_.type == "block":
                        function_block = child_

                assert function_name is not None, "Function name"

                parent_names = self.get_node_ns(node_)
                if parent_names:
                    function_name = f"{'.'.join(parent_names)}.{function_name}"

                function_markers: List[RelationMarkerType] = []
                block_comment = None
                if (
                    function_block is not None
                    and len(function_block.children) > 0
                    and function_block.children[0].type
                    == "expression_statement"
                ):
                    if len(function_block.children[0].children) > 0:
                        if (
                            function_block.children[0].children[0].type
                            == "string"
                        ):
                            block_comment = function_block.children[0].children[
                                0
                            ]
                            # String contains of three parts:
                            # (string_start string_content string_end)
                            string_content = block_comment.children[1]
                            assert string_content.text is not None

                            block_comment_text = string_content.text.decode(
                                "utf-8"
                            )
                            source_node = MarkerParser.parse(
                                block_comment_text,
                                node_.start_point[0] + 1,
                                node_.end_point[0] + 1,
                                string_content.start_point[0] + 1,
                                function_name,
                            )
                            for marker_ in source_node.markers:
                                if isinstance(marker_, FunctionRangeMarker):
                                    function_range_marker_processor(
                                        marker_, parse_context
                                    )
                                    traceability_info.markers.append(marker_)
                                    function_markers.append(marker_)

                # FIXME: This look more complex than needed but can't make mypy happy.
                cursor_: Optional[Node] = node_
                while cursor_ is not None and (cursor_ := cursor_.parent):
                    if cursor_ == map_function_to_node[functions_stack[-1]]:
                        break
                    if cursor_.type == "function_definition":
                        functions_stack.pop()
                        assert len(functions_stack) > 0
                else:
                    # This is counterintuitive:
                    # The top-level functions don't have the top-level module set
                    # as their parent, so in this branch, we simply clear the whole
                    # function stack, leaving the top module only.
                    functions_stack = functions_stack[:1]

                new_function = Function(
                    parent=traceability_info,
                    name=function_name,
                    display_name=function_name,
                    line_begin=node_.range.start_point[0] + 1,
                    line_end=node_.range.end_point[0] + 1,
                    child_functions=[],
                    # Python functions do not need to track markers.
                    markers=[],
                    attributes=set(),
                )
                map_function_to_node[new_function] = node_

                parent_function = functions_stack[-1]

                parent_function.child_functions.append(new_function)
                functions_stack.append(new_function)
                traceability_info.functions.append(new_function)

                traceability_info.ng_map_names_to_markers[function_name] = (
                    function_markers
                )
            elif node_.type == "comment":
                assert node_.text is not None, (
                    f"Comment without a text: {node_}"
                )

                node_text_string = node_.text.decode("utf8")

                source_node = MarkerParser.parse(
                    node_text_string,
                    node_.start_point[0] + 1,
                    node_.end_point[0] + 1,
                    node_.start_point[0] + 1,
                    None,
                )
                for marker_ in source_node.markers:
                    if isinstance(marker_, RangeMarker) and (
                        range_marker := marker_
                    ):
                        range_marker_processor(range_marker, parse_context)
                    elif isinstance(marker_, LineMarker) and (
                        line_marker := marker_
                    ):
                        line_marker_processor(line_marker, parse_context)
                    else:
                        pass
            else:
                pass

        assert (
            functions_stack[0].name == "module"
            or functions_stack[0].name == "translation_unit"
        )

        source_file_traceability_info_processor(
            traceability_info, parse_context
        )

        return traceability_info

    def read_from_file(self, file_path: str) -> SourceFileTraceabilityInfo:
        with open(file_path, "rb") as file:
            sdoc_content = file.read()
            sdoc = self.read(sdoc_content, file_path=file_path)
            return sdoc

    @staticmethod
    def get_node_ns(node: Node) -> Sequence[str]:
        """
        Walk up from node to find enclosing class and function names.

        Handles nested functions, methods, and classes.
        """
        parent_scopes: List[str] = []
        cursor: Optional[Node] = node

        while cursor is not None:
            if cursor.type in ("class_definition", "function_definition"):
                # Look for the identifier child (i.e., the name).
                name_node = next(
                    (
                        child
                        for child in cursor.children
                        if child.type == "identifier"
                    ),
                    None,
                )
                if name_node and name_node.text:
                    parent_scopes.insert(0, name_node.text.decode("utf-8"))
            cursor = cursor.parent

        # The array now contains the "fully qualified" node name,
        # we want to return the namespace, so don't return the last part.
        return parent_scopes[:-1]



VALID_AFTER_INLINE_MARKUP = " \t\n-.,:;!?\\/'\")]}>"


def escape_str_after_inline_markup(str_after_inline_markup: str) -> str:
    # Ensure inline markup recognition rule #7
    # https://docutils.sourceforge.io/docs/ref/rst/restructuredtext.html#inline-markup-recognition-rules
    if (
        str_after_inline_markup
        and str_after_inline_markup[0] not in VALID_AFTER_INLINE_MARKUP
    ):
        return f"\\{str_after_inline_markup}"
    return str_after_inline_markup



"""
@relation(SDOC-SRS-142, scope=file)
"""

from typing import Generator, Optional, Tuple, Union

from tree_sitter import Node, Tree


def traverse_tree(tree: Tree) -> Generator[Node, None, None]:
    cursor = tree.walk()

    visited_children = False
    while True:
        if not visited_children:
            if cursor.node is not None:
                yield cursor.node
            if not cursor.goto_first_child():
                visited_children = True
        elif cursor.goto_next_sibling():
            visited_children = False
        elif not cursor.goto_parent():
            break


def ts_find_child_node_by_type(
    node: Node, node_type: Union[str, Tuple[str, ...], str]
) -> Optional[Node]:
    node_types: Tuple[str, ...] = (
        node_type if isinstance(node_type, tuple) else (node_type,)
    )
    for child_ in node.children:
        if child_.type in node_types:
            return child_
    return None


def ts_find_child_nodes_by_type(
    node: Node, node_type: str
) -> Generator[Node, None, None]:
    for child_ in node.children:
        if child_.type == node_type:
            yield child_
